/*****************************************************************************
 * Copyright (c) Future Technology Devices International 2014
 * propriety of Future Technology devices International.
 *
 * Software License Agreement
 *
 * This code is provided as an example only and is not guaranteed by FTDI. 
 * FTDI accept no responsibility for any issues resulting from its use. 
 * The developer of the final application incorporating any parts of this 
 * sample project is responsible for ensuring its safe and correct operation 
 * and for any consequences resulting from its use.
 *****************************************************************************/
/**
 * @file		VM800P4350_RS232_Master.ino
 * @brief		Sketch to demonstrate FT800 graphics and RS232 module
 *				To be used in conjunction with sketch VM800P4350_RS232_Slave.ino
 *				Tested platform version: Arduino 1.0.4 and later
 * @version		0.1.0
 * @date		2014/12/19
 *
 */

// Arduino standard includes ...
#include "SPI.h"
#include "Wire.h"

// Platform specific includes ...
#include "FT_VM800P43_50.h" // FT800
#include "FT_VI800A_232U.h" // RS232

// Uncomment to enable serial debug output ...
//#define SERIAL_ENABLE

// Global object for FT800 implementation ...
FT800IMPL_SPI FTImpl(FT_CS_PIN,FT_PDN_PIN,FT_INT_PIN);

// Global object for RS232 implementation ...
FT_VI800A VI800A;

// Global for SD card object
FT_SD FtSd(FT_SD_CSPIN);

// Global for SD card status - holds error values during initialization (0 means no error, otherwise an error).
FT_SDStatus sd_present; 

// Global SD file object for file accessing
FT_SDFile sdFileHandler;

// ================================================================================================
// VARIABLES/ENUMERATIONS [RS232]
// ================================================================================================
typedef enum
{
	BAUD_001200 = 0,
	BAUD_001800,
	BAUD_002000,
	BAUD_002400,
	BAUD_003600,
	BAUD_004800,
	BAUD_007200,
	BAUD_009600,
	BAUD_019200,
	BAUD_038400,
	BAUD_056000,
	BAUD_057600,
	BAUD_115200,
	BAUD_230400,
	BAUD_460800,
	BAUD_921600,
	BAUD_TOTAL
} baudrates_t;

typedef enum
{
	STATE_TX_IDLE = 0,
	STATE_TX_DATA,
	STATE_RX_IDLE,
	STATE_RX_DATA
} states_t;

typedef enum
{
	// Command list must be exactly the same (including order) for master and slave code!
	COMMAND_WAITING_FOR_REPLY = 0,
	COMMAND_HELLO_THERE,
	COMMAND_HELLO_THERE_CONFIRMED,
	COMMAND_BAUDRATE_CHANGE,
	COMMAND_BAUDRATE_CHANGE_CONFIRMED,
	COMMAND_IMAGE_START,
	COMMAND_IMAGE_START_CONFIRMED,
	COMMAND_IMAGE_SLICE,
	COMMAND_IMAGE_SLICE_CONFIRMED,
	COMMAND_IMAGE_DONE,
	COMMAND_IMAGE_DONE_CONFIRMED,
	COMMAND_SELECT_BAUDRATE,
	COMMAND_WAIT_FOR_A_WHILE,
	COMMAND_END_OF_COMMANDS
} commands_t;

uint32_t baudrates[BAUD_TOTAL] =
{
	  1200UL,
	  1800UL,
	  2000UL,
	  2400UL,
	  3600UL,
	  4800UL,
	  7200UL,
	  9600UL,
	 19200UL,
	 38400UL,
	 56000UL,
	 57600UL,
	115200UL,
	230400UL,
	460800UL,
	921600UL
};

char * baudrateText[BAUD_TOTAL] =
{
	  "1200",
	  "1800",
	  "2000",
	  "2400",
	  "3600",
	  "4800",
	  "7200",
	  "9600",
	 "19200",
	 "38400",
	 "56000",
	 "57600",
	"115200",
	"230400",
	"460800",
	"921600"
};

#define  COMMAND_BLOCK_SIZE 4U
#define  NO_PARAMETER       0U
#define  NO_DATA            0U

uint8_t  dataRxState     = STATE_RX_IDLE;
uint8_t  dataRxCommand   = COMMAND_WAITING_FOR_REPLY;
uint16_t dataRxParameter = 0;
uint16_t dataRxDataSize  = 0;
uint8_t  dataRxReady     = true;
uint16_t dataRxIndex;
uint16_t dataRxBytes;
uint8_t  byteStoreRx[32];

uint8_t  dataTxState     = STATE_TX_IDLE;
uint8_t  dataTxCommand   = COMMAND_HELLO_THERE;
uint16_t dataTxParameter = 0;
uint16_t dataTxDataSize  = 0;
uint8_t  dataTxReady     = false;
uint16_t dataTxIndex;
uint16_t dataTxBytes;
uint8_t  byteStoreTx[512];

uint8_t  baudrateAlter   = false;
uint8_t  baudrateIndex   = BAUD_038400;

// ================================================================================================
// VARIABLES/ENUMERATIONS [SWIPE LIST DISPLAY AND CONTROL]
// ================================================================================================
// Structure used to hold information about swipe lists (position, size etc.) ...
typedef struct
{
	uint8_t  Layout; // direction of list (SWIPE_VERTICAL, SWIPE_HORIZONTAL, SWIPE_BOTH_WAYS)
	uint8_t  Number; // total number of items
	uint8_t  Select; // currently selected item
	uint8_t  Xentry; // X spacing of items
	uint8_t  Yentry; // Y spacing of items
	int16_t  Xspeed; // X speed of list (due to a swipe)
	int16_t  Yspeed; // Y speed of list (due to a swipe)
	int16_t  Xstart; // current display offset of list
	int16_t  Ystart; // current display offset of list
	int16_t  Xcoord; // active area X coordinate (top left)
	int16_t  Ycoord; // active area Y coordinate (top left)
	int16_t  Xwidth; // active area X width
	int16_t  Ywidth; // active area Y width
	uint8_t  Xitems; // number of items to display horizontally
	uint8_t  Yitems; // number of items to display vertically
	uint8_t  Limits; // full wrap around or stop at limits
	uint8_t  Active; // Whether or not list is active
} swipeList_t;

// Swipe list display function prototypes ...
typedef void (* openptr) (void);
typedef	void (* itemptr) (int16_t, int16_t, uint16_t, uint16_t, uint8_t, uint8_t);
typedef void (* shutptr) (void);

// Swipe list display functions structure (user supplied functions used to render swipe lists)...
typedef struct
{
	openptr openFunc; // performs any initial operations before item rendering
	itemptr itemFunc; // performs rendering of an item at supplied position
	shutptr shutFunc; // performs any final operations after item rendering
} swipeDraw_t;

// list of user defined swipe lists ...
typedef enum
{
	SWIPE_LIST_IMAGE = 0,
	SWIPE_LIST_TOTAL
} swipeListUser_t;

// Swipe list directions and limits ...
#define SWIPE_BOTH_WAYS  0x00
#define SWIPE_VERTICAL   0x01
#define SWIPE_HORIZONTAL 0x02
#define SWIPE_LIMIT      0x00
#define SWIPE_WRAPS      0x01

// Swipe list bits ...
#define SWIPE_NULL_TAG     255
#define SWIPE_REFRESH_RATE  20
#define SWIPE_DRAG_PIXELS    3
#define SWIPE_TOUCH_DRAG     0
#define SWIPE_TOUCH_SELECT   1

// Swipe list item size for IMAGES swipe list...
#define SWIPE_ITEM_XSIZ_IMAGE 116
#define SWIPE_ITEM_YSIZ_IMAGE  91

swipeList_t swipeItemList[SWIPE_LIST_TOTAL];
swipeDraw_t swipeItemDraw[SWIPE_LIST_TOTAL];

// Swipe lists - holds value of REG_TOUCH_DIRECT_XY register ...
uint32_t touchRegsValueXY;

// Swipe lists - tags being checked and tracked ...
uint8_t swipeCheckingTag = SWIPE_NULL_TAG;
uint8_t swipeTrackingTag = SWIPE_NULL_TAG;

// Swipe lists - coordinates of initial touch point ...
int16_t swipeInitialXcrd;
int16_t swipeInitialYcrd;

// Swipe lists - coordinates of current touch point ...
int16_t swipeCurrentXcrd;
int16_t swipeCurrentYcrd;

// Swipe lists - coordinates of previous touch point ...
int16_t touchEarlierXcrd;
int16_t touchEarlierYcrd;

// Swipe lists - difference between initial and current touch points ...
int16_t swipeDiffersXcrd = 0;
int16_t swipeDiffersYcrd = 0;

// Swipe lists - whether a drag or selection ...
uint8_t swipeTypeOfTouch = SWIPE_TOUCH_SELECT;

// ================================================================================================
// VARIABLES/ENUMERATIONS [IMAGES LOADED FROM SD CARD]
// ================================================================================================
#define FT_RAM_CMD 0x00108000
#define IMAGE_TOTAL 10
#define IMAGE_FULL 1
#define IMAGE_HALF 0.5

// Bitmap structure ...
typedef struct Bitmap_header
{
	uint8_t  Format;
	uint16_t Width;
	uint16_t Height;
	uint16_t Stride;
	uint32_t Offset;
} Bitmap_header_t;

// Image info ...
Bitmap_header_t imageInfo[IMAGE_TOTAL] = 
{
	// format, width, height, stride, offset
	{FT_RGB565, 1, 1, 1, 0},
	{FT_RGB565, 1, 1, 1, 0},
	{FT_RGB565, 1, 1, 1, 0},
	{FT_RGB565, 1, 1, 1, 0},
	{FT_RGB565, 1, 1, 1, 0},
	{FT_RGB565, 1, 1, 1, 0},
	{FT_RGB565, 1, 1, 1, 0},
	{FT_RGB565, 1, 1, 1, 0},
	{FT_RGB565, 1, 1, 1, 0},
	{FT_RGB565, 1, 1, 1, 0},
};

// Image filenames ...
char imageName[IMAGE_TOTAL][12] =
{
	"LandB01.jpg",
	"LandB02.jpg",
	"LandB03.jpg",
	"LandB04.jpg",
	"LandB05.jpg",
	"LandB06.jpg",
	"LandB07.jpg",
	"LandB08.jpg",
	"LandB09.jpg",
	"LandB10.jpg"
};

Bitmap_header_t * headerInfo;
uint32_t ramStore = FT_RAM_G;
uint8_t imagePick = 0;

// Used during transfer of image ...
uint16_t imageTxStride   = 0; // Bytes per line of image
uint16_t imageTxTotal    = 0; // Total lines in image
uint16_t imageTxSlice    = 0; // Current line of image

// Used to animate image being transferred ...
uint8_t  animateScanOn   = false;
uint8_t  animateCycles   = 0;
int8_t   stepperCycles   = 1;

// ================================================================================================
// VARIABLES/ENUMERATIONS [VARIOUS]
// ================================================================================================
// Used to decrement 'msTickTock...' counters at a millisecond(ish) rate ...
uint32_t msTimeCountWas;
uint32_t msTimeCountNow;

// Used to update display and track touches at the millisecond rate specified ...
#define DISPLAY_REFRESH_RATE 20
uint16_t msTickTockDisplay = DISPLAY_REFRESH_RATE; 

// Used to animate item selected in a swipe list (glowing rectangle) ...
uint8_t  animateSelect = 0;
int8_t   stepperSelect = 1;

// used to perform various actions when decremented to 0 ...
uint16_t msTickTockGeneral =   0;                  // general purpose use

// ################################################################################################
// # BEGINNING OF CODE
// ################################################################################################

// ================================================================================================
// Timer routine
// ================================================================================================
void milliTimer()
{
	msTimeCountWas = msTimeCountNow;
	msTimeCountNow = millis();

	// update count downs if a millisecond (or so) has passed ...
	if (msTimeCountNow != msTimeCountWas)
	{
		if (msTickTockDisplay)
		{
			msTickTockDisplay--;
			
			// Update display if count down reaches 0 ...
			if (msTickTockDisplay == 0)
			{
				// reset timer ...
				msTickTockDisplay = DISPLAY_REFRESH_RATE;
				// update display ...
				renderDisplay();
			}
		}

		if (msTickTockGeneral)
		{
			msTickTockGeneral--;
		}
	}
}

// ================================================================================================
// Swipe list core routines ...
// ================================================================================================
// ------------------------------------------------------------------------------------------------
// Swipe list setup (offsets, speeds and item selected all default to 0) ...
// ------------------------------------------------------------------------------------------------
void swipeSetupList(uint8_t list, uint8_t type, uint8_t line, uint8_t item, uint8_t xgrd, uint8_t ygrd, int16_t xcrd, int16_t ycrd, int16_t xsiz, int16_t ysiz, uint8_t wrap)
{
	swipeItemList[list].Layout = type;
	swipeItemList[list].Number = item;
	swipeItemList[list].Select = 0;
	swipeItemList[list].Xentry = xgrd;
	swipeItemList[list].Yentry = ygrd;
	swipeItemList[list].Xspeed = 0;
	swipeItemList[list].Yspeed = 0;
	swipeItemList[list].Xstart = 0;
	swipeItemList[list].Ystart = 0;
	swipeItemList[list].Xcoord = xcrd;
	swipeItemList[list].Ycoord = ycrd;
	swipeItemList[list].Xwidth = xsiz;
	swipeItemList[list].Ywidth = ysiz;
	swipeItemList[list].Limits = wrap;
	swipeItemList[list].Active = true;

	// 'line' specifies either the number of columns or rows depending on the layout chosen...
	switch (type)
	{
		case SWIPE_BOTH_WAYS:  // 'line' = column total, rows calculated, can move any direction
		case SWIPE_VERTICAL:   // 'line' = column total, rows calculated, can only move vertically
			swipeItemList[list].Xitems = line;
			swipeItemList[list].Yitems = (item + line - 1) / line;
			break;
		case SWIPE_HORIZONTAL: // 'line' = row total, columns calculated, can only move horizontally
			swipeItemList[list].Xitems = (item + line - 1) / line;
			swipeItemList[list].Yitems = line;
			break;
	}
}

// ------------------------------------------------------------------------------------------------
// Swipe list offset limiting routines (ensures X/Y offsets stay within relevant limits)
// ------------------------------------------------------------------------------------------------
int16_t swipeLimitX(int16_t xoff, uint8_t xtag)
{
	int16_t xmax = swipeItemList[xtag].Xitems * swipeItemList[xtag].Xentry;

	// Determine limits according to limit type ...
	if (swipeItemList[xtag].Limits == SWIPE_WRAPS)
	{
		// Ensure offset wraps around ...
		xoff = (xoff + xmax) % xmax;
	}
	else
	{
		// Ensure offset is kept within limits ...
		xmax =  xmax - swipeItemList[xtag].Xwidth;
		xoff = (xoff < 0)? 0 : (xoff > xmax)? xmax : xoff;
	}

	return xoff;
}

int16_t swipeLimitY(int16_t yoff, uint8_t ytag)
{
	int16_t ymax = swipeItemList[ytag].Yitems * swipeItemList[ytag].Yentry;

	// Determine limits according to limit type ...
	if (swipeItemList[ytag].Limits == SWIPE_WRAPS)
	{
		// Ensure offset wraps around ...
		yoff = (yoff + ymax) % ymax;
	}
	else
	{
		// Ensure offset is kept within limits ...
		ymax -= swipeItemList[ytag].Ywidth;
		yoff  = (yoff < 0)? 0 : (yoff > ymax)? ymax : yoff;
	}

	return yoff;
}

// ------------------------------------------------------------------------------------------------
// Swipe list touch handling routine (handles user selection and dragging actions)
// ------------------------------------------------------------------------------------------------
void swipeTouchHandler()
{
	static int16_t swipePrecedeXcrd;
	static int16_t swipePrecedeYcrd;
	int16_t xmin, ymin;
	int16_t xmax, ymax;
	int16_t xsiz, ysiz;
	int16_t xoff, yoff;

	uint8_t tag;
	uint8_t item;
	uint8_t xitm, yitm;

	// Get current touch coordinates and time
	touchRegsValueXY = FTImpl.Read32(REG_TOUCH_DIRECT_XY);
	swipeCheckingTag = SWIPE_NULL_TAG;

	// Determine swipe area being touched (if any) and set tag appropriately ...
	if ( (touchRegsValueXY & 0x80000000) == 0)
	{
		// Record preceding touch coordinates (used to calculate direction and speed of swipe) ...
		swipePrecedeXcrd = swipeCurrentXcrd;
		swipePrecedeYcrd = swipeCurrentYcrd;

		// Extract current touch coordinates (need to invert touch Y to match screen Y) ...
		swipeCurrentXcrd =      (((touchRegsValueXY >> 16) & 0x3FF) * 0.468750); // Scale to the range 0-479
		swipeCurrentYcrd = 271 -(((touchRegsValueXY      ) & 0x3FF) * 0.265625); // Scale to the range 0-271

		// Check to see if touch has occurred in any of the swipe areas ...
		for (tag = 0; tag < SWIPE_LIST_TOTAL; tag++)
		{
			if (swipeItemList[tag].Active == true)
			{
				xmin = swipeItemList[tag].Xcoord;
				ymin = swipeItemList[tag].Ycoord;
				xmax = swipeItemList[tag].Xwidth + xmin;
				ymax = swipeItemList[tag].Ywidth + ymin;
				
				// Once a swipe area is being tracked it will continue to be tracked until the touch point
				// moves outside of the touch area or the touch finishes, even if swipe areas overlap and
				// are listed earlier in the swipe areas to be checked.
				if ( (swipeCurrentXcrd >= xmin) && (swipeCurrentXcrd <= xmax) && (swipeCurrentYcrd >= ymin) && (swipeCurrentYcrd <= ymax) )
				{
					// Variable will be set to track the first swipe area which the touch position is bounded by, but will
					// be ignored below if we are already tracking another swipe area.
					swipeCheckingTag = tag;
					break;
				}
			}
		}
	}

	// Determine whether to start, continue or finish tracking ...
	if ((swipeCheckingTag == SWIPE_NULL_TAG) && (swipeTrackingTag != SWIPE_NULL_TAG))
	{
		// FINISH TRACKING ...
		xsiz = swipeItemList[swipeTrackingTag].Xentry;
		ysiz = swipeItemList[swipeTrackingTag].Yentry;
		xmax = swipeItemList[swipeTrackingTag].Xitems * xsiz;
		ymax = swipeItemList[swipeTrackingTag].Yitems * ysiz;

		// Perform appropriate action depending on whether a select or drag action ..
		if (swipeTypeOfTouch == SWIPE_TOUCH_SELECT)
		{
			// Determine coordinate of point touched with respect to current offsets ...
			xoff = swipeCurrentXcrd - swipeItemList[swipeTrackingTag].Xcoord;
			yoff = swipeCurrentYcrd - swipeItemList[swipeTrackingTag].Ycoord;

			// Determine item being selected ...
			xitm = ((swipeItemList[swipeTrackingTag].Xstart + xoff) % xmax) / xsiz;
			yitm = ((swipeItemList[swipeTrackingTag].Ystart + yoff) % ymax) / ysiz;
			item = (swipeItemList[swipeTrackingTag].Xitems * yitm) + xitm;

			// Only update selected item if valid (multi column/row lists may not be fully populated) ...
			if (item < swipeItemList[swipeTrackingTag].Number)
			{
				swipeItemList[swipeTrackingTag].Select = item;
			}
		}
		else
		{
			// Update start positions of list display and keep within range ...
			swipeItemList[swipeTrackingTag].Xstart = swipeLimitX(swipeItemList[swipeTrackingTag].Xstart + swipeDiffersXcrd, swipeTrackingTag);
			swipeItemList[swipeTrackingTag].Ystart = swipeLimitY(swipeItemList[swipeTrackingTag].Ystart + swipeDiffersYcrd, swipeTrackingTag);
			// Determine speed/direction of movement ...
			swipeItemList[swipeTrackingTag].Xspeed = (swipeItemList[swipeTrackingTag].Layout == SWIPE_VERTICAL)?   0 : swipePrecedeXcrd - swipeCurrentXcrd;
			swipeItemList[swipeTrackingTag].Yspeed = (swipeItemList[swipeTrackingTag].Layout == SWIPE_HORIZONTAL)? 0 : swipePrecedeYcrd - swipeCurrentYcrd;
		}

		// Reset to initial values ...
		swipeTrackingTag = SWIPE_NULL_TAG;
		swipeTypeOfTouch = SWIPE_TOUCH_SELECT;
		swipeDiffersXcrd = 0;
		swipeDiffersYcrd = 0;
	}
	else if ((swipeCheckingTag != SWIPE_NULL_TAG) && (swipeTrackingTag == SWIPE_NULL_TAG))
	{
		// START TRACKING ...
		// Make note of tag to track and coordinates ...
		swipeTrackingTag = swipeCheckingTag;
		swipeInitialXcrd = swipeCurrentXcrd;
		swipeInitialYcrd = swipeCurrentYcrd;

		// Initialise speed/direction for tag area touched...
		swipeItemList[swipeTrackingTag].Xspeed = 0;
		swipeItemList[swipeTrackingTag].Yspeed = 0;

		// Initialise last two touch coordinates (used to calculate speed/direction of swipe) ...
		swipePrecedeXcrd = swipeCurrentXcrd;
		swipePrecedeYcrd = swipeCurrentYcrd;
	}
	else if ((swipeCheckingTag != SWIPE_NULL_TAG) && (swipeTrackingTag != SWIPE_NULL_TAG))
	{
		// CONTINUE TRACKING ...
		// Determine difference between initial and current touch positions ...
		swipeDiffersXcrd = (swipeItemList[swipeTrackingTag].Layout == SWIPE_VERTICAL)?   0 : swipeInitialXcrd - swipeCurrentXcrd;
		swipeDiffersYcrd = (swipeItemList[swipeTrackingTag].Layout == SWIPE_HORIZONTAL)? 0 : swipeInitialYcrd - swipeCurrentYcrd;

		// Only becomes a drag touch if difference between initial and current touch
		// positions exceeds the specified number of pixels ...
		if ((abs(swipeDiffersXcrd) > SWIPE_DRAG_PIXELS) || (abs(swipeDiffersYcrd) > SWIPE_DRAG_PIXELS))
		{
			// Change to a drag touch and track movement ...
			swipeTypeOfTouch = SWIPE_TOUCH_DRAG;
		}
	}
}

// ------------------------------------------------------------------------------------------------
// Swipe list display generation routine (uses user supplied functions to do actual drawing)
// ------------------------------------------------------------------------------------------------
void swipeDisplayHandler()
{
	int16_t xcrd, ycrd; // Coordinates to display item
	int16_t xsiz, ysiz; // Size of items to be displayed
	int16_t xoff, yoff; // Coordinates of display offset
	int16_t xtmp;
	uint8_t tag;        // List being displayed
	uint8_t xitm, yitm; // Used to determine item at specific coordinates
	uint8_t item, itmp;

	// Update item selection animation ...
	animateSelect += stepperSelect;
	if ((animateSelect == 0) || (animateSelect == 15))
	{
		stepperSelect = -stepperSelect;
	}

	// Render all active swipe lists ...
	for (tag = 0; tag < SWIPE_LIST_TOTAL; tag++)
	{
		if (swipeItemList[tag].Active == true)
		{
			xoff = swipeItemList[tag].Xstart;
			yoff = swipeItemList[tag].Ystart;
			xsiz = swipeItemList[tag].Xentry;
			ysiz = swipeItemList[tag].Yentry;

			// Alter display position if touch is being tracked for this swipe list and list is being dragged ... 
			if ((swipeTrackingTag == tag) && (swipeTypeOfTouch == SWIPE_TOUCH_DRAG))
			{
				xoff = swipeLimitX(xoff + swipeDiffersXcrd, tag);
				yoff = swipeLimitY(yoff + swipeDiffersYcrd, tag);
			}

			// Determine position of first item to be rendered ...
			xcrd = swipeItemList[tag].Xcoord - (xoff % xsiz);
			ycrd = swipeItemList[tag].Ycoord - (yoff % ysiz);
			xtmp = xcrd;

			// Determine first item to be rendered ...
			xitm = (xoff / xsiz);
			yitm = (yoff / ysiz);
			itmp = xitm;
			
			// Save graphics context ...
			FTImpl.SaveContext();
			
			// Set clip rectangle to swipe list display area ...
			FTImpl.ScissorXY(  swipeItemList[tag].Xcoord, swipeItemList[tag].Ycoord);
			FTImpl.ScissorSize(swipeItemList[tag].Xwidth, swipeItemList[tag].Ywidth);
			
			// Perform any item specific operation required before rendering items (user supplied routine) ...
			swipeItemDraw[tag].openFunc();

			// Render items, automatically determining how many are required ...
			while (ycrd < (swipeItemList[tag].Ycoord + swipeItemList[tag].Ywidth))
			{
				while (xcrd < (swipeItemList[tag].Xcoord + swipeItemList[tag].Xwidth))
				{
					// Determine next item to be displayed ...
					item = (yitm * swipeItemList[tag].Xitems) + xitm;

					// Only display if a valid item ...
					if (item < swipeItemList[tag].Number)
					{
						// Display item at coordinates specified (user supplied routine) ...
						swipeItemDraw[tag].itemFunc(xcrd, ycrd, xsiz, ysiz, item, swipeItemList[tag].Select);
					}

					// Update position and item to be displayed (X direction) ...
					xcrd += xsiz;
					xitm  = (xitm + 1) % swipeItemList[tag].Xitems;
				}

				// Update position and item to be displayed (Y direction) ...
				ycrd += ysiz;
				yitm  = (yitm + 1) % swipeItemList[tag].Yitems;
				xcrd  = xtmp;
				xitm  = itmp;			
			}

			// Perform any item specific operation required after rendering items (user supplied routine) ...
			swipeItemDraw[tag].shutFunc();

			// Restore graphics context ...
			FTImpl.RestoreContext();

			// Reduce X/Y speed ...
			swipeItemList[tag].Xspeed += (swipeItemList[tag].Xspeed < 0)? 1 : (swipeItemList[tag].Xspeed > 0)? -1 : 0;
			swipeItemList[tag].Yspeed += (swipeItemList[tag].Yspeed < 0)? 1 : (swipeItemList[tag].Yspeed > 0)? -1 : 0;

			// Update X/Y offsets (keeping within limits) ...
			swipeItemList[tag].Xstart  = swipeLimitX(swipeItemList[tag].Xstart + swipeItemList[tag].Xspeed, tag);
			swipeItemList[tag].Ystart  = swipeLimitY(swipeItemList[tag].Ystart + swipeItemList[tag].Yspeed, tag);
		}
	}
}
// ================================================================================================
// Swipe list display rendering routines for the IMAGE list
// ================================================================================================
void swipeRenderOpen_Image()
{
	// Load default identity matrix, change scale and then set identity matrix ...
	FTImpl.Cmd_LoadIdentity();
	FTImpl.Cmd_Scale(IMAGE_HALF * 65536, IMAGE_HALF * 65536);
	FTImpl.Cmd_SetMatrix();
	// Set line width ...
	FTImpl.LineWidth(4 * 16);
}

void swipeRenderItem_Image(int16_t xcrd, int16_t ycrd, uint16_t xsiz, uint16_t ysiz, uint8_t item, uint8_t pick)
{
	// Open section ...
	FTImpl.Begin(FT_RECTS);

	// Highlight item if currently selected ...
	if (item == pick)
	{
		// Draw rectangle to indicate bitmap is selected ...
		FTImpl.ColorRGB(0xFF - (animateSelect * 8), 0xD7 - (animateSelect * 8), 0x00);
		FTImpl.LineWidth(8 * 16);
		FTImpl.Vertex2f((xcrd + 8       ) * 16, (ycrd + 8       ) * 16); // Top left
		FTImpl.Vertex2f((xcrd + xsiz - 9) * 16, (ycrd + ysiz - 9) * 16); // Bottom right
		FTImpl.LineWidth(4 * 16);
	}

	// Draw background rectangle to display bitmap on ...
	FTImpl.ColorRGB(0x00, 0x00, 0x00);
	FTImpl.Vertex2f((xcrd + 8       ) * 16, (ycrd + 8       ) * 16); // Top left
	FTImpl.Vertex2f((xcrd + xsiz - 9) * 16, (ycrd + ysiz - 9) * 16); // Bottom right

	// Shut section ...
	FTImpl.End();

	// Draw image at half size ...
	if (imageInfo[item].Stride != 0)
	{
		// Open section ...
		FTImpl.Begin(FT_BITMAPS);
		// Set the source address of bitmap ..
		FTImpl.BitmapSource(imageInfo[item].Offset);
		// Set the layout of the bitmap ...
		FTImpl.BitmapLayout(imageInfo[item].Format, imageInfo[item].Stride, imageInfo[item].Height);
		// Set the bitmap size ...
		FTImpl.BitmapSize(FT_BILINEAR, FT_BORDER, FT_BORDER, IMAGE_HALF * imageInfo[item].Width, IMAGE_HALF * imageInfo[item].Height);
		// Set colour ...
		FTImpl.ColorRGB(0xFF, 0xFF, 0xFF);
		// Display bitmap ...
		FTImpl.Vertex2f((xcrd + (xsiz - imageInfo[item].Width / 2) / 2) * 16, (ycrd + (ysiz - imageInfo[item].Height / 2) / 2) * 16);
		// Shut section ...
		FTImpl.End();
	}
}

void swipeRenderShut_Image()
{
	// nothing to do!
}

// ================================================================================================
// Routine to generate display list
// ================================================================================================
void renderDisplay()
{
	uint16_t imageWidth  = headerInfo->Width;
	uint16_t imageHeight = headerInfo->Height;
	uint16_t imageStride = headerInfo->Stride;
	int32_t  xcrd;
	int32_t  ycrd;

	// ------------------------------------------------------------------------
	// Track all touches on swipe lists ...
	// ------------------------------------------------------------------------
	swipeTouchHandler();
	
	// ------------------------------------------------------------------------
	// Backdrop for image transfer display (max image size of 200 x 150) ...
	// ------------------------------------------------------------------------
	FTImpl.DLStart(); // Note DLStart and DLEnd are helper apis, Cmd_DLStart() and Display() can also be utilized.

	// Draw entire background using a couple of gradient fills ...
	FTImpl.ScissorXY(     0,   0);
	FTImpl.ScissorSize( 480, 136);
	FTImpl.Cmd_Gradient(  0,   0, 0x202020,   0, 136, 0xFFFFFF);
	FTImpl.ScissorXY(     0, 136);
	FTImpl.ScissorSize( 480, 136);
	FTImpl.Cmd_Gradient(  0, 136, 0xFFFFFF,   0, 272, 0x202020);

	// Reset scissor rectangle to full screen ...
	FTImpl.ScissorXY(     0,   0);
	FTImpl.ScissorSize( 480, 272);

	// Open section ...
	FTImpl.Begin(FT_RECTS);
	FTImpl.ColorA(128);
	FTImpl.LineWidth(4 * 16);

	// Semi-transparent rectangle for background of text display (left side) ...
	FTImpl.ColorRGB(0x00, 0x00, 0x30);
	FTImpl.Vertex2f(  8 * 16,   8 * 16);
	FTImpl.Vertex2f(127 * 16, 160 * 16);
	// Semi-transparent rectangle for background of text display (right side) ...
	FTImpl.Vertex2f(352 * 16,   8 * 16);
	FTImpl.Vertex2f(471 * 16, 160 * 16);
	// Semi-transparent rectangle for background of animated transfer display ...
	FTImpl.ColorRGB(0x00, 0x00, 0x00);
	FTImpl.Vertex2f(140 * 16,   8 * 16);
	FTImpl.Vertex2f(339 * 16, 160 * 16);
	// Semi-transparent rectangle for background of image swipe list area...
	FTImpl.ColorRGB(0x00, 0x40, 0x40);
	FTImpl.Vertex2f(  8 * 16, 172 * 16);
	FTImpl.Vertex2f(471 * 16, 262 * 16);
	FTImpl.ColorRGB(0xFF, 0xFF, 0xFF);

	// Shut section ...
	FTImpl.End();

	if (imageStride == 0)
	{
		// ------------------------------------------------------------------------
		// Information about image that failed to load ...
		// ------------------------------------------------------------------------
		FTImpl.Cmd_Text(140, 100, 28, 0, "Failed to load image");
		FTImpl.Cmd_Text(140, 120, 30, 0, imageName[imagePick]);
	}
	else
	{
		// ------------------------------------------------------------------------
		// Information about how the swipe list works ...
		// ------------------------------------------------------------------------
		FTImpl.Cmd_Text(    8,   8, 26, 0, "Touch and drag");
		FTImpl.Cmd_Text(    8,  28, 26, 0, "left / right to view");
		FTImpl.Cmd_Text(    8,  48, 26, 0, "items in the list.");
		FTImpl.Cmd_Text(    8,  76, 26, 0, "Tap on an item to");
		FTImpl.Cmd_Text(    8,  96, 26, 0, "select it.");
		FTImpl.Cmd_Text(    8, 124, 26, 0, "Flick left / right to");
		FTImpl.Cmd_Text(    8, 144, 26, 0, "scroll list quickly.");

		// ------------------------------------------------------------------------
		// Information about image being transferred ...
		// ------------------------------------------------------------------------
		FTImpl.Cmd_Text(  352,   8, 26, 0, "FILENAME:");
		FTImpl.Cmd_Text(  362,  28, 26, 0, imageName[imagePick]);
		FTImpl.Cmd_Text(  352,  48, 26, 0, "IMAGE SIZE (x y):");
		FTImpl.Cmd_Number(362,  68, 26, 0, imageWidth);
		FTImpl.Cmd_Number(422,  68, 26, 0, imageHeight);
		FTImpl.Cmd_Text(  352,  88, 26, 0, "IMAGE BYTES:");
		FTImpl.Cmd_Number(362, 108, 26, 0, imageStride * imageHeight);
		FTImpl.Cmd_Text(  352, 128, 26, 0, "BAUDRATE:");
		FTImpl.Cmd_Text(  362, 148, 26, 0, baudrateText[baudrateIndex]);
		FTImpl.ColorA(255);
		
		// ------------------------------------------------------------------------
		// Animation of image being transferred (emulates a scanner light) ...
		// ------------------------------------------------------------------------
		FTImpl.Begin(FT_BITMAPS);
		// Set the bitmap size ...
		FTImpl.BitmapSize(FT_BILINEAR, FT_BORDER, FT_BORDER, IMAGE_FULL * imageWidth, IMAGE_FULL * imageHeight);
		// Set the source address of bitmap ..
		FTImpl.BitmapSource(headerInfo->Offset);
		// Set the layout of the bitmap ...
		FTImpl.BitmapLayout(headerInfo->Format, imageStride, imageHeight);
	
		// Set scissor rectangle to image size and offset display of image
		// to only display parts of image still to be transmitted ...
		xcrd = 240 - (imageWidth  / 2);
		ycrd =  83 - (imageHeight / 2);
		FTImpl.ScissorXY(xcrd, ycrd);
		FTImpl.ScissorSize(imageWidth, imageHeight);
		FTImpl.Vertex2f(xcrd * 16, (ycrd - imageTxSlice) * 16);
		FTImpl.End();

		// Set scissor rectangle to full screen ...
		FTImpl.ScissorXY(     0,   0);
		FTImpl.ScissorSize( 480, 272);

		// Display scanning light if transmitting image ...
		if (animateScanOn)
		{
			// Update scanning light animation ...
			animateCycles += stepperCycles;
			if ((animateCycles == 0) || (animateCycles == 5))
			{
				stepperCycles = -stepperCycles;
			}

			// Display transparent thick white line to show 'glow' of scanning light ...
			FTImpl.Begin(FT_LINES);
			FTImpl.ColorA(64);
			FTImpl.LineWidth(32);
			FTImpl.Vertex2f( xcrd * 16,               ycrd * 16);
			FTImpl.Vertex2f((xcrd + imageWidth) * 16, ycrd * 16);
			FTImpl.ColorA(255);

			// Display solid thin white line once in a while to 'flash' scanning light ...
			if ((animateCycles == 0) || (animateCycles == 2))
			{
				FTImpl.LineWidth(16);
				FTImpl.Vertex2f( xcrd * 16,               ycrd * 16);
				FTImpl.Vertex2f((xcrd + imageWidth) * 16, ycrd * 16);
			}

			FTImpl.End();
		}
	}

	// ------------------------------------------------------------------------
	// Render all active swipe lists ...
	// ------------------------------------------------------------------------
	swipeDisplayHandler();
	
	// ------------------------------------------------------------------------
	// End of display list
	// ------------------------------------------------------------------------
	FTImpl.DLEnd();  // End the display list
	FTImpl.Finish(); // Render the display list and wait for the completion of the DL
}

// ================================================================================================
// Core routine for transferring data from FT800 graphics ram
// ================================================================================================
void copyDataFromRamG(uint16_t slice)
{
	// number of bytes per slice (line of image) ...
	uint16_t stride = headerInfo->Stride;
	// calculate offset into FT800 graphics ram (start of image + slice required) ...
	uint32_t offset = headerInfo->Offset + (slice * stride);
	// copy 'stride' number of bytes from FT800 graphics ram into Tx storage area ...
	FTImpl.Read(offset, (uint8_t *)&byteStoreTx[COMMAND_BLOCK_SIZE], stride);
}

// ================================================================================================
// Core routines for receiving data via RS232
// ================================================================================================
void commandBytesHandlerRx()
{
	// Check to see if we are ready to receive data ...
	if (dataRxReady)
	{	
		// Check to see if there is any data in the receive FIFO ...
		if(VI800A.isDataInReceiver())
		{
			// Receive next byte of data ...
			byteStoreRx[dataRxIndex++] = VI800A.Read();
		
			// Until the required number of bytes have been received ...
			if (dataRxIndex == dataRxBytes)
			{
				dataRxReady = false;
				commandBlockHandlerRx();
			}
		}
	}
}

void commandBlockHandlerRx()
{
	// Retrieve command, parameter and size of additional data ...
	dataRxCommand   = byteStoreRx[0];
	dataRxParameter = byteStoreRx[1];
	dataRxDataSize  = byteStoreRx[2];
	dataRxDataSize |= byteStoreRx[3] << 0x08;

	// Determine if additional data has been supplied and received ...
	if (dataRxDataSize == 0)
	{
		// No additional data supplied so wait for next command ...
		dataRxState = STATE_RX_IDLE;
	}
	else
	{
		if (dataRxBytes == COMMAND_BLOCK_SIZE)
		{
			// Additional data still to be received so continue receiving ...
			dataRxReady  = true;
			dataRxBytes += dataRxDataSize;
		}
		else
		{
			// Additional data already received so wait for next command ...
			dataRxState = STATE_RX_IDLE;
		}
	}
}

void commandStateHandlerRx()
{
	switch (dataRxState)
	{
		case STATE_RX_IDLE: // Determine what to do next ...
			switch (dataRxCommand)
			{
				case COMMAND_WAITING_FOR_REPLY:
					// Wait for command block data ...
					dataRxReady = true;
					dataRxIndex = 0;
					dataRxBytes = COMMAND_BLOCK_SIZE;
					dataRxState = STATE_RX_DATA;
					break;

				case COMMAND_BAUDRATE_CHANGE:
					// Set baudrate to supplied value ...
					baudrateIndex = dataRxParameter;

					// Switch to the new baudrate now ...
					VI800A.SetBaudRate(baudrates[baudrateIndex]);

					// Wait for command block data (we don't send a response)...
					dataRxCommand = COMMAND_WAITING_FOR_REPLY;
					break;

				case COMMAND_HELLO_THERE_CONFIRMED:
				case COMMAND_IMAGE_START_CONFIRMED:
				case COMMAND_IMAGE_SLICE_CONFIRMED:
				case COMMAND_IMAGE_DONE_CONFIRMED:
					if (imagePick != swipeItemList[SWIPE_LIST_IMAGE].Select)
					{
						// Ignore command response as we have selected a new
						// image so begin transmitting the new image ...
						imagePick     = swipeItemList[SWIPE_LIST_IMAGE].Select;
						dataTxCommand = COMMAND_IMAGE_START;
					}
					else
					{
						// Pass command block to Tx handler ...
						dataTxCommand   = dataRxCommand;
						dataTxParameter = dataRxParameter;
						dataTxDataSize  = dataRxDataSize;
					}
					
					// Wait for command block data ...
					dataRxCommand   = COMMAND_WAITING_FOR_REPLY;
					break;
			}
			break;
			
		case STATE_RX_DATA: // Retrieve data ...
			commandBytesHandlerRx();
			break;
	}
}

// ================================================================================================
// Core routines for transmitting data via RS232
// ================================================================================================
void commandBytesHandlerTx()
{
	// Check to see if we are ready to transmit data ...
	if (dataTxReady)
	{
		// Check to see if there is enough room in the transmit FIFO ...
		if (VI800A.GetTransmitFIFOFreeSpace())
		{
			// Transmit next byte of data ...
			VI800A.Write(byteStoreTx[dataTxIndex++]);
			
			// Until the required number of bytes have been transmitted ...
			if (--dataTxBytes == 0)
			{
				dataTxReady = false;
				dataTxState = STATE_TX_IDLE;
			}
		}
	}
}

void commandBlockHandlerTx(uint8_t command, uint8_t parameter, uint16_t datasize)
{
	// Store command, parameter and size of data to be sent
	// <data should be set up before calling this routine>
	byteStoreTx[0] = command;
	byteStoreTx[1] = parameter;
	byteStoreTx[2] = (datasize        ) & 0xFF;
	byteStoreTx[3] = (datasize >> 0x08) & 0xFF;
	
	// Transmit buffer containing command, parameter and extra data (if supplied) ...
	datasize   += COMMAND_BLOCK_SIZE;
	dataTxReady = true;
	dataTxIndex = 0;
	dataTxBytes = datasize;
	dataTxState = STATE_TX_DATA;

	// Wait for reply ...
	dataTxCommand = COMMAND_WAITING_FOR_REPLY;
}

void commandStateHandlerTx()
{
	uint8_t * p;
	uint8_t   i;

	switch (dataTxState)
	{
		case STATE_TX_IDLE: // Handling of commands
			switch (dataTxCommand)
			{
				case COMMAND_WAITING_FOR_REPLY:
					// Commands will be automatically passed on by Rx handler for processing!
					break;

				case COMMAND_HELLO_THERE:
					// Send command to establish communications and wait for reply ...
					commandBlockHandlerTx(COMMAND_HELLO_THERE, NO_PARAMETER, NO_DATA);
					break;

				case COMMAND_HELLO_THERE_CONFIRMED:
					// Communications established, start transmitting images ...
					dataTxCommand = COMMAND_IMAGE_START;
					break;

				case COMMAND_IMAGE_START:
					headerInfo = &imageInfo[imagePick];

					if (headerInfo->Stride == 0)
					{
						// Will display 'failed to load' message ...
						dataTxCommand = COMMAND_WAIT_FOR_A_WHILE;
						msTickTockGeneral = 2000;
					}
					else
					{
						// Copy image header data to transmit buffer ...
						p = (uint8_t *)headerInfo;

						for (i = 0; i < 11; i++)
						{
							byteStoreTx[COMMAND_BLOCK_SIZE+i] = *p++;
						}

						// Send command to initiate transfer of image data and wait for reply ...
						commandBlockHandlerTx(COMMAND_IMAGE_START, 0, 11);
					}
					break;
			
				case COMMAND_IMAGE_START_CONFIRMED:
					// Initiate image data transfer ...
					dataTxCommand = COMMAND_IMAGE_SLICE;
					imageTxSlice  = 0;
					imageTxTotal  = headerInfo->Height;
					imageTxStride = headerInfo->Stride;
					animateScanOn = true;
					break;
					
				case COMMAND_IMAGE_SLICE:
					// Copy image slice data from FT800 ram to transmit buffer ...
					copyDataFromRamG(imageTxSlice);
					// Send command and image slice data ...
					commandBlockHandlerTx(COMMAND_IMAGE_SLICE, imageTxSlice, imageTxStride);
					break;

				case COMMAND_IMAGE_SLICE_CONFIRMED:
					// Select next slice to be sent ...
					imageTxSlice++;

					// Image data has been received, so send next slice or return to idle state ..
					dataTxCommand = (imageTxSlice < imageTxTotal)? COMMAND_IMAGE_SLICE : COMMAND_IMAGE_DONE;
					break;

				case COMMAND_IMAGE_DONE:
					// Send command to conclude image transfer and wait for reply ...
					commandBlockHandlerTx(COMMAND_IMAGE_DONE, NO_PARAMETER, NO_DATA);
					break;

				case COMMAND_IMAGE_DONE_CONFIRMED:
					animateScanOn = false;

					// Repeat over and over for now ...
					dataTxCommand = COMMAND_WAIT_FOR_A_WHILE;
					msTickTockGeneral = 1000;
					break;

				case COMMAND_WAIT_FOR_A_WHILE:
					if (msTickTockGeneral == 0)
					{
						dataTxCommand = COMMAND_HELLO_THERE;
					}
					break;
			}
			break;
			
		case STATE_TX_DATA: // Handling of command block transfer via RS232 (command + parameter + data)
			commandBytesHandlerTx();
			break;
	}
}

// ================================================================================================
// Routine to check for presence of SD card
// ================================================================================================
void CheckStorageDevicePrecence()
{
	// Display an on-screen message if the storage device is not found 
	while (sd_present)
	{
		FTImpl.DLStart();     
		FTImpl.Clear(1, 1, 1);
		FTImpl.ColorRGB(255, 255, 255);
		FTImpl.Cmd_Text(FT_DISPLAYWIDTH >> 1, FT_DISPLAYHEIGHT >> 1, 29, FT_OPT_CENTER, "STORAGE DEVICE NOT FOUND");
		FTImpl.DLEnd();
		FTImpl.Finish();
		delay(5000);
	}
}

// ================================================================================================
// Routines to load JPEG data into ram, transfer it to the FT800 and inflate it into the FT800 GRAM
// ================================================================================================
void loadJpegBytes(FT_SDFile &sdFileRef)
{
	uint16_t byteRead;

	while (sdFileRef.Offset < sdFileRef.Size)
	{
		// determine number of bytes to read this time (maximum of 512 bytes at a time) ...
		byteRead = min(512, sdFileRef.Size - sdFileRef.Offset);
		// force 32-bit alignment ...
		byteRead = (byteRead + 3) & ~3;
		// read next sector into buffer ...
		sdFileRef.ReadSector(byteStoreTx);
		// copy data continuously into command memory ...
		FTImpl.WriteCmd(byteStoreTx, byteRead);
	}
}

void loadJpegImage(uint8_t index)
{
	uint32_t imageAddress;
	uint32_t imageWidth;
	uint32_t imageHeight;
	uint16_t results;

	// wait until FT800 is ready ...
	while ( FTImpl.Read16(REG_CMD_WRITE) != FTImpl.Read16(REG_CMD_READ) );
	
	if ( FtSd.OpenFile(sdFileHandler, imageName[index]) )
	{
		// Failed to open/find file so set image size to 0x0 pixels ...
		imageWidth   = 0;
		imageHeight  = 0;
	}
	else
	{  
		// load JPEG image
		FTImpl.WriteCmd(CMD_LOADIMAGE);
		// destination address
		FTImpl.WriteCmd(ramStore);
		// options
		FTImpl.WriteCmd(2); // OPT_NODL = 2 | OPT_RGB565 = 0;
		// load data
		loadJpegBytes(sdFileHandler);
		
		// #######################################################################################
		// ACCESSING RESULTS OF A COMMAND SENT TO THE FT800
		// #######################################################################################
		// 1) Wait until CMD_LOADIMAGE command has completed ...
		while ( FTImpl.Read16(REG_CMD_WRITE) != FTImpl.Read16(REG_CMD_READ) );
		
		// 2) Note the current value of REG_CMD_WRITE and store in variable 'results'
		results = FTImpl.Read16(REG_CMD_WRITE);

		// 3) Send the command and parameters 
		FTImpl.Cmd_GetProps(imageAddress, imageWidth, imageHeight);
		
		// 4) Read result(s) - 1st result at 'results+4', 2nd at 'results+8' etc ..
		imageWidth   = FTImpl.Read32(FT_RAM_CMD + results +  8);
		imageHeight  = FTImpl.Read32(FT_RAM_CMD + results + 12);
	}  

	// store image information (format is always FT_RGB565) ...
	imageInfo[index].Format = FT_RGB565;
	imageInfo[index].Width  = imageWidth;
	imageInfo[index].Height = imageHeight;
	imageInfo[index].Stride = imageWidth * 2;
	imageInfo[index].Offset = ramStore;
	
	// update address to store next image at ...
	ramStore += imageHeight * imageWidth * 2;
}

// ================================================================================================
// Routine to boot up FT800 module, verify hardware and configure display/audio pins
//(returns 0 in case of success and 1 in case of failure)
// ================================================================================================
int16_t BootupConfigure()
{
	uint32_t chipid = 0;
	// Configure the display to the WQVGA ...
	FTImpl.Init(FT_DISPLAY_RESOLUTION);
	
	delay(20);
	chipid = FTImpl.Read32(FT_ROM_CHIPID);

	// Identify the chip ...
	if(FT800_CHIPID != chipid)
	{
		#ifdef SERIAL_ENABLE
		Serial.print("Error in chip id read ");
		Serial.println(chipid,HEX);
		#endif
		return 1;
	}

	// Set the Display & audio pins
	FTImpl.SetDisplayEnablePin(FT_DISPENABLE_PIN);
	FTImpl.SetAudioEnablePin(FT_AUDIOENABLE_PIN); 
	FTImpl.DisplayOn(); 
	FTImpl.AudioOn();  			
	return 0;
}

// ================================================================================================
// Routine to calibrate touch screen
// ================================================================================================
void Calibrate()
{  
	// ------------------------------------------------------------------------
	// Below code demonstrates the usage of calibrate function.
	// - The 'Calibrate' function waits until the user presses all three dots.
	// - Only way to exit this api is to reset the coprocessor bit.
	// ------------------------------------------------------------------------

	// Construct the display list with grey as background color, display informative
	// string "Please Tap on the dot" followed by inbuilt calibration command ...
	FTImpl.DLStart();
	FTImpl.ClearColorRGB(64,64,64);
	FTImpl.Clear(1,1,1);    
	FTImpl.ColorRGB(0xff, 0xff, 0xff);
	FTImpl.Cmd_Text((FT_DISPLAYWIDTH/2), (FT_DISPLAYHEIGHT/2), 27, FT_OPT_CENTER, "Please Tap on the dot");
	FTImpl.Cmd_Calibrate(0);
	
	// Wait for the completion of calibration(either finish can be used for flush and check can be used)
	FTImpl.Finish();
}

// ================================================================================================
// Routine to boot up the module and render display
// ================================================================================================
void setup()
{
	uint8_t load;
	
	// Initialize serial print related functionality
	#ifdef SERIAL_ENABLE
	Serial.begin(9600);
	Serial.println("-- Start Application --");	
	#endif

	// Initialize the SD object (screen error message can only be displayed when the FT800 is ready) ...
	sd_present = FtSd.Init();

	if (BootupConfigure())
	{
		// Error case - do not do any thing
		#ifdef SERIAL_ENABLE
		Serial.println("Error in: BootupConfigure()");
		#endif
	}
	else
	{
		// -----------------------------------------------------------------------------------------
		// RS232 SETUP
		// -----------------------------------------------------------------------------------------
		// Set data format ... same as calling VI800A.Init(FT_232U_CS)
		VI800A.Init(FT_232U_CS, baudrates[baudrateIndex], FT_No_Parity, FT_One_Stop_Bit, FT_Eight_Data_Bits);	
		VI800A.SetBaudRate(baudrates[baudrateIndex]);

		// Use hardware flow control but no FIFO trigger
		VI800A.SetFlowControl(FT_HW_Flow, FT_Use_No_RX_Trigger);	

		// Enable FIFO (FIFO is enabled implicitly if FIFO trigger levels are specified) ...
		VI800A.EnableFIFO(FT_Enable);
		VI800A.clearFIFO();
		delay(100);
		
		CheckStorageDevicePrecence();

		// Load images ...
		for (load = 0; load < IMAGE_TOTAL; load++)
		{
			loadJpegImage(load);
		}
		
		// Calibrate touch screen ...
		Calibrate();

		// Point to first image ...
		headerInfo = &imageInfo[imagePick];
		
		// set up IMAGE swipe list (movement = SWIPE_BOTH_WAYS, SWIPE_VERTICAL or SWIPE_HORIZONTAL) ...
		swipeSetupList(SWIPE_LIST_IMAGE, SWIPE_HORIZONTAL, 1, IMAGE_TOTAL, SWIPE_ITEM_XSIZ_IMAGE, SWIPE_ITEM_YSIZ_IMAGE,      8, 172, 464, SWIPE_ITEM_YSIZ_IMAGE, SWIPE_WRAPS);
		// set up IMAGE swipe list rendering functions ...
		swipeItemDraw[SWIPE_LIST_IMAGE].openFunc = &swipeRenderOpen_Image;
		swipeItemDraw[SWIPE_LIST_IMAGE].itemFunc = &swipeRenderItem_Image;
		swipeItemDraw[SWIPE_LIST_IMAGE].shutFunc = &swipeRenderShut_Image;

		// Initialise item selected for swipe lists ...
		swipeItemList[SWIPE_LIST_IMAGE].Select = imagePick;

		while (1)
		{
			// Handle RS232 transmitting ...
			commandStateHandlerTx();
			// Handle RS232 receiving ...
			commandStateHandlerRx();
			// Check countdown timers and perform defined operations if required ...
			milliTimer();
		}
	}

	FTImpl.Exit();
	#ifdef SERIAL_ENABLE
	Serial.println("-- End Application --");
	#endif
}


void loop()
{
	// nothing in loop
}

// ################################################################################################
// END OF CODE
// ################################################################################################
