/*****************************************************************************
* Copyright (c) Future Technology Devices International 2014
* propriety of Future Technology devices International.
*
* Software License Agreement
*
* This code is provided as an example only and is not guaranteed by FTDI. 
* FTDI accept no responsibility for any issues resulting from its use. 
* The developer of the final application incorporating any parts of this 
* sample project is responsible for ensuring its safe and correct operation 
* and for any consequences resulting from its use.
*****************************************************************************/
/**
* @file                           FT_VI800A.h
* @brief                          Contains the necessary functionalities to utilize the VI800A series add-on modules.
								  Tested platform version: Arduino 1.0.4 and later
								  Dependencies: Arduino standard SPI library
* @version                        1.1.0
* @date                           2014/Oct/01
*

VERSION SIGNIFICANCE:
1.0.0 - 2014/Oct/1 - Initial release
1.1.0 - 2015/Jan/07 - Updated enumeration structure names with the library name prepended at the front  of the name to prevent library conflicts. NOTE: This version is not backward compatible with previous version(s). 

*/
#ifndef _FT_VI800A_H_
#define _FT_VI800A_H_

#define FT_VI800A_MAJOR				(1)
#define FT_VI800A_MINOR				(1)
#define FT_VI800A_BUILD				(0)


//used for APIs to enable/disable functionalities
typedef enum FT_VI800A_OPTIONS{
	FT_VI800A_Enable=1,
	FT_VI800A_Disable=2,
}FT_VI800A_OPTIONS;

//used by the SetFlowControl function to specify the desired flow control.
typedef enum FT_VI800A_Flow_Type{
	FT_VI800A_No_Flow=0,
	FT_VI800A_SW_Flow=1,
	FT_VI800A_HW_Flow=2,
	FT_VI800A_RS485_Flow=3,
}FT_VI800A_Flow_Type;


typedef enum FT_VI800A_HW_FLOW_OPTIONS{
	FT_VI800A_Enable_Auto_CTS=128,
	FT_VI800A_Enable_Auto_RTS=64,
}FT_VI800A_HW_FLOW_OPTIONS;

//modes for the RS485 flow control
typedef enum FT_VI800A_RS485_FLOW_OPTIONS{
	FT_VI800A_RS485_Multidrop_Mode=0,
	FT_VI800A_Auto_RS485_Address_Detection_Mode=4,  
}FT_VI800A_RS485_FLOW_OPTIONS;

//the options are used for setting the software and hardware flow control FIFO trigger levels.  FT_VI800A_Use_No_RX_Trigger option specifies no trigger level will be used. FT_VI800A_Use_TCR_RX_Trigger option uses the TCR register which has a finer trigger level but using the TCR register will prevent the usage of MSR(give modem status) and SPR(working register for user to read and write) registers.  FT_VI800A_Use_FCR_RX_Trigger option uses the FCR register which only offers 4 trigger levels but the MSR and SPR registers are accessible. If the  FT_VI800A_Use_FCR_RX_Trigger option is used then only the halt level is needed because the next lower trigger level is automatically used for the resume level.
typedef enum FT_VI800A_Flow_Trigger{
	FT_VI800A_Use_No_RX_Trigger=0,
	FT_VI800A_Use_TCR_RX_Trigger=1,
	FT_VI800A_Use_FCR_RX_Trigger=2,
}FT_VI800A_Flow_Trigger;


//the status of the interrupt in a prioritized order.
typedef enum FT_VI800A_Interrupts_Status{
	FT_VI800A_No_Interrupts = 255,//An interrupt is pending only if the first bit is a logic 0
	FT_VI800A_Receiver_Line_Status_Error = 6, //receive line status error, check the FT_VI800A_FIFO_Status table
	FT_VI800A_Receiver_Time_Out_Interrupt = 12, //Receiver time-out interrupt, if there is no more incoming data after 4 character time frame has passed.
	FT_VI800A_RHR_Interrupt=4, //RHR interrupt
	FT_VI800A_THR_Interrupt=2, //THR interrupt
	FT_VI800A_Modem_Interrupt=0, //modem interrupt status is reflected by reading the MSR register
	FT_VI800A_Input_Pin_State_Change=48,//input pin change of state, reflected by reading the IOState register
	FT_VI800A_Received_Xoff_Signal=16, //received xoff signal/ special character
	FT_VI800A_CTS_RTS_Inactive=32, //cts, rts change of state from active (low) to inactive (high)
}FT_VI800A_Interrupts_Status;


//The FCR RX FIFO trigger level is the pre-defined trigger levels used for software and hardware flow controls, if FIFO is enabled.  A finer level can be specified in the TCR register. However, the utilization of TCR and TLR, set by MCR[2], will prevent the access to MSR(modem status) and SPR(scratch pad) registers.
typedef enum FT_VI800A_Preset_RX_FIFO_Trigger_Level{
	FT_VI800A_RX_Eight_Characters=0,
	FT_VI800A_RX_Sixteen_Characters=64,
	FT_VI800A_RX_Fifty_Six_Characters=128,
	FT_VI800A_RX_Sixty_Characters=192,
}FT_VI800A_Preset_RX_FIFO_Trigger_Level;

//The FCR TX FIFO trigger level is the the pre-defined trigger level used for software and hardware flow control, if FIFO is enabled.  A finer level can be specified in the TCR register. However, the utilization of TCR and TLR, set by MCR[2], will prevent the access to MSR(modem status) and SPR(scratch pad) registers.
typedef enum FT_VI800A_Preset_TX_FIFO_Trigger_Level{
	FT_VI800A_TX_Eight_Characters=0,
	FT_VI800A_TX_Sixteen_Characters=16,
	FT_VI800A_TX_Thirty_Two_Characters=32,
	FT_VI800A_TX_Fifty_Six_Characters=48,
}FT_VI800A_Preset_TX_FIFO_Trigger_Level;

//These errors only apply to the FIFO
typedef enum FT_VI800A_FIFO_Status{
	FT_VI800A_FIFO_OK=0,
	FT_VI800A_Overrun_Error=2, //overrun error has occurred
	FT_VI800A_Parity_Error=4, //Parity error in data being read from RX FIFO
	FT_VI800A_Framing_Error=8, //Framing error occurred in data being read from RX FIFO, received daa did not have a valid stop bit.
	FT_VI800A_Break_Interrupt=16, //a break condition occurred and associated character is 0x00, which the RX was LOW for one character time frame.
}FT_VI800A_FIFO_Status;

//device interrupts
typedef enum FT_VI800A_Enableable_Interrupts{
	FT_VI800A_Enable_All_Interrupts=239,
	FT_VI800A_CTS_Interrupt=128,
	FT_VI800A_RTS_Interrupt=64,
	FT_VI800A_Xoff_Interrupt=32,
	FT_VI800A_Modem_Status_Interrupt=8,
	FT_VI800A_Receive_Line_Status_Interrupt=4,
	FT_VI800A_Transmit_Holding_Register_Interrupt=2,
	FT_VI800A_Receive_Holding_Register_Interrupt=1,
	FT_VI800A_Enable_No_Interrupts=0,	
}FT_VI800A_Enableable_Interrupts;

//error enums here,
typedef enum FT_VI800A_Status{
	FT_VI800A_OK=0,
	FT_VI800A_Out_Of_Order_Setup_Error = 1,  //some features need specific steps to enable.
	FT_VI800A_Hardware_Error=2,  //undefined fatal error
	FT_VI800A_Configuration_Error=4, //
	FT_VI800A_Feature_Not_Enabled=8, 
	FT_VI800A_Feature_FIFO_Time_Out=16,
	FT_VI800A_Parameter_Value_Error=32,
	FT_VI800A_FIFO_Is_Full=64,
	FT_VI800A_FIFO_Is_Empty=128,
}FT_VI800A_Status;

/* 
Enhanced software flow control combination.  NOTE: User should only use one of two groups: 
	Group 1: one of FT_VI800A_SW_Flow_Combination[0:3] bitwise OR'ed with one of FT_VI800A_SW_Flow_Combination[4:6]
	Group 2: one of FT_VI800A_SW_Flow_Combination[7:10]
*/
typedef enum FT_VI800A_SW_Flow_Combination{
	FT_VI800A_NO_TX_FLOW = 0x00,  //no transmit flow control
	FT_VI800A_TX_XON1_XOFF1 = 0x08, //transmit xon1, xoff1
	FT_VI800A_TX_XON2_XOFF2 = 0x04, //transmit xon2, xoff2
	FT_VI800A_TX_XON1_XON2_XOFF1_XOFF2 = 0x0c, //transmit xon1 and xon2, xoff1 and xoff2
	FT_VI800A_NO_RX_FLOW = 0x00, //no receive flow control
	FT_VI800A_RX_XON1_XOFF1 = 0x02, //receiver compares xon1, xoff1
	FT_VI800A_RX_XON2_XOFF2 = 0x01, //receiver compares xon2, xoff2
	FT_VI800A_TX_XON1_XOFF1_RX_XON1_OR_XON2_XOFF1_OR_XOFF2 = 0x0b,//transmit xon1, xoff1 and receiver compares xon1 or xon2, xoff1 or xoff2
	FT_VI800A_TX_XON2_XOFF2_RX_XON1_OR_XON2_XOFF1_OR_XOFF2 = 0x07,//transmit xon2, xoff2 and receiver compares xon1 or xon2, xoff1 or xoff2
	FT_VI800A_TX_XON1_XON2_XOFF1_XOFF2_RX_XON1_XON2_XOFF1_XOFF2 = 0x0f,//transmit xon1 and xon2, xoff1 and xoff2 and receiver compares xon1 and xon2, xoff1 and xoff2
	FT_VI800A_NO_TX_FLOW_RX_XON1_XON2_XOFF1_XOFF2 = 0x03,//no transmit flow control but receiver compares xon1 and xon2, xoff1 and xff2
}FT_VI800A_SW_Flow_Combination;

//user look at the table to pick the desired option
typedef enum FT_VI800A_Stop_Bit_Length{
	FT_VI800A_One_Stop_Bit=0,  //this options is used on reset
	FT_VI800A_One_And_A_Half_Stop_Bit=4,
	FT_VI800A_Two_Stop_Bits=4,
}FT_VI800A_Stop_Bit_Length;

typedef enum FT_VI800A_Word_Length{
	FT_VI800A_Five_Data_Bits=0, //5 bits
	FT_VI800A_Six_Data_Bits=1,  //6 bits
	FT_VI800A_Seven_Data_Bits=2, //7 bits
	FT_VI800A_Eight_Data_Bits=3, //8 bits
}FT_VI800A_Word_Length;

typedef enum FT_VI800A_Parity_Type{
	FT_VI800A_No_Parity=0,			 
	FT_VI800A_Odd_Parity=8,         
	FT_VI800A_Even_Parity=24,       
	FT_VI800A_Forced_Parity_One=40, 
	FT_VI800A_Forced_Parity_Zero=56,
}FT_VI800A_Parity_Type;



class FT_VI800A{
	private:
	
	uint8_t sspin;
	uint8_t FCR;
	
	FT_VI800A_Status WriteToRegister(uint8_t address, uint8_t value);	
	
	void ResetFlowFunctions();
	
	uint8_t ReadRegister(uint8_t address);
	
	public:
	
	void GetVersion(uint8_t &Major, uint8_t &Minor, uint8_t &Build);
	
	FT_VI800A_Status Init(uint8_t SSpin);		
	
	FT_VI800A_Status Init(uint8_t SSpin, uint32_t BaudRate, FT_VI800A_Parity_Type parityType, FT_VI800A_Stop_Bit_Length stopBitLength, FT_VI800A_Word_Length wordLength);
	
	uint8_t Read();
	
	FT_VI800A_Status Read(uint8_t* array, uint8_t maxBytesToRead, uint32_t &numOfBytesRead);
	
	uint8_t Write(uint8_t data);
	
	uint8_t Write(uint8_t* array, uint32_t size);
	
	void SetDataFormat(FT_VI800A_Parity_Type parityType, FT_VI800A_Stop_Bit_Length stopBitLength, FT_VI800A_Word_Length wordLength);

	void SetBaudRate(uint32_t baudrate);
	
	uint8_t GetTransmitFIFOFreeSpace();
	
	uint8_t GetReceiveFIFOFillLevel();
	
	void SetInterrupts(uint8_t interrupts);

	void EnableSleepMode(FT_VI800A_OPTIONS option);
	
	void SWReset();
	
	uint8_t GetInterrupts(void);
	
	FT_VI800A_Status SetFlowControl(FT_VI800A_Flow_Type FlowControlType, uint8_t param1, uint8_t param2, uint8_t param3, uint8_t param4, uint8_t param5, uint8_t param6, uint8_t param7, uint8_t param8);
		
	void SetInternalLoopBack(FT_VI800A_OPTIONS option);
	
	void EnableFIFO(FT_VI800A_OPTIONS option);
	
	void EnableTransmitter(FT_VI800A_OPTIONS option);
	
	void EnableReceiver(FT_VI800A_OPTIONS option);
	      
	uint8_t GetReceiveFIFOStatus(void);

	void SendXON1SpecialCharacter();
	
	void SendXON2SpecialCharacter();
	
	void SendXOFF1SpecialCharacter();
	
	void SendXOFF2SpecialCharacter();

	void EnableRTSOutput(FT_VI800A_OPTIONS option);
	
	bool isCTSActive(void);
	
	uint8_t ReadScratchPad(void);
	
	void WriteToScratchPad(uint8_t value);
	
	FT_VI800A_Status SetFIFOInterruptTriggerLevel(uint8_t TransmitFIFOInterruptLevel, uint8_t ReceiveFIFOInterruptLevel);
	
	void SetPresetFIFOInterruptTriggerLevel(FT_VI800A_Preset_RX_FIFO_Trigger_Level ReceiveFIFOInterruptLevel, FT_VI800A_Preset_TX_FIFO_Trigger_Level TransmitFIFOInterruptLevel);


	void SetIOPinDirection(uint8_t Direction);
	
	uint8_t GetIOPinDirection(void);
	
	void SetIOPinState(uint8_t State);
	
	//void EnableModemPins(void);
	
	void StartTransmitLineBreak(void);
	
	void StopTransmitLineBreak(void);
	
	bool isDataInReceiver(void);
	
	bool isTransmitFIFOandShiftRegisterEmpty(void);
	
	bool isTransmitFIFOEmpty(void);
	
	void clearFIFO(void);
	
	void Exit(void);
	};
	

	
	/*
	Description: write an array of n number of bytes to the transmit FIFO.  There is only 64 bytes in the FIFO,if FIFO is enabled, and the actual amount of data transfered to the FIFO is dictated by the amount of the free space available at the time of calling this function.  Note: Data transferred to the transmit FIFO doesn't mean the data is actually sent.  The transmitter has to be enabled and a call to isTransmitFIFOandShiftRegisterEmpty() and/or isTransmitFIFOEmpty() can confirm whether the data in the transmit FIFO have actually been sent.
	
	Parameters:
		uint8_t* address - an 8-bit address of the array.
		
		uint32_t length - the amount of data to be transfered, the maximum this function can be transferred 64 bytes.
		
	Return:
		
		An 8-bit number represents the number of transfered bytes, value between 0 to 64.
	*/
	uint8_t FT_VI800A::Write(uint8_t* address, uint32_t length) 
	{
		uint8_t amountToTransfer = 0;
		uint8_t FIFOfreespace = GetTransmitFIFOFreeSpace();

		if(FIFOfreespace==0 || length==0)
			return 0;
		amountToTransfer = FIFOfreespace>length ? length : FIFOfreespace;
		digitalWrite(sspin,LOW);
		SPI.transfer(FT_VI800A_THR);
		for(uint8_t i = 0; i < amountToTransfer; i++){
			SPI.transfer(address[i]);
		}
		digitalWrite(sspin,HIGH);
		return amountToTransfer;
	}
	
	/*
	Description: Send a single byte of data to the transmit FIFO.  Note: Data transferred to the transmit FIFO doesn't mean the byte has actually been sent.  The transmitter has to be enabled and a call to isTransmitFIFOandShiftRegisterEmpty() and/or isTransmitFIFOEmpty() can confirm whether the byte has actually been sent.
	
	Parameters:
		uint8_t data - An 8-bit data byte.
		
	Return:
		A value of either 0 or 1.  1 means the the data byte is transferred to transmit FIFO, 0 means the data was not transmitted to the FIFO because it is full.
	*/	
	uint8_t FT_VI800A::Write(uint8_t data){	
		if(GetTransmitFIFOFreeSpace()){
			digitalWrite(sspin,LOW);
			SPI.transfer(FT_VI800A_THR);
			SPI.transfer(data);
			digitalWrite(sspin,HIGH);
			return 1;
		}
		return 0;
	}

	
	/*
	Description: Internal function to set a register.
	
	Parameters:
		uint8_t address - The address of the register.
		
		uint8_t data - The value of the register, all registers are 8-bit in size.
		
	Return:
		FT_VI800A_Status status
	*/
	FT_VI800A_Status FT_VI800A::WriteToRegister(uint8_t address, uint8_t data){
		digitalWrite(sspin,LOW);
		SPI.transfer(address);
		SPI.transfer(data);
		digitalWrite(sspin,HIGH);
		return FT_VI800A_OK;
	}
	

	 
	/*
	Description: Internal function to read a register.
	
	Parameters:
		uint8_t address - The address of the register to be read.
		
	Return:
		An 8-bit value of the register.
	*/
	uint8_t FT_VI800A::ReadRegister(uint8_t address){
		address = (address | 0x80);
		digitalWrite(sspin,LOW);
		SPI.transfer(address);   
		uint8_t data = SPI.transfer(0x00);
		digitalWrite(sspin,HIGH);
		return data;  
	}
	

	/*
	Description: Get the current version of the library.
	
	Parameters: 
		uint8_t &Major - an 8-bit variable that will get updated with the major version number.
		
		uint8_t &Minor - an 8-bit variable that will get updated with the minor version number.
		
		uint8_t &Build - an 8-bit variable that will get updated with the build version number.
		
	Return:
		None.
	*/
	void FT_VI800A::GetVersion(uint8_t &Major, uint8_t &Minor, uint8_t &Build)
	{
		Major = FT_VI800A_MAJOR;
		Minor = FT_VI800A_MINOR;
		Build = FT_VI800A_BUILD;
	}
	
	/*
	Description: Read one byte from the FIFO.  User is expected to make sure there is data in the FIFO before calling this function by using the isDataInReceiver() or GetReceiveFIFOFillLevel() function.  Reading an empty FIFO will get junk data.
	
	Parameters:
		None.
		
	Return:
		An 8-bit value from the top of the FIFO.
	*/
	uint8_t FT_VI800A::Read(){
		return ReadRegister(FT_VI800A_RHR);
	}
	
	
	/*
	Description: Read n number of bytes from the FIFO.  This function checks the receive FIFO fill level and no more than 64 bytes will be read.
	
	Parameters:
		uint8_t* array - An 8-bit pointer to an array to hold the data.
		
		uint8_t maxBytesToRead - The number of bytes to read.  There are 64 bytes in the FIFO, that is if the FIFO is enabled, otherwise a maximum of 1 byte will be read.
		
		uint32_t &numOfBytesRead - A variable to hold the actually number of bytes read from the receive FIFO.  The updated value is between 0 to 64.  
		
	Return: 
		FT_VI800A_Status - One entry in the FT_VI800A_Status table.
	*/
	FT_VI800A_Status FT_VI800A::Read(uint8_t* array, uint8_t maxBytesToRead, uint32_t &numOfBytesRead){
		uint8_t FreeSpace = GetReceiveFIFOFillLevel();
		uint8_t address = (FT_VI800A_RHR | 0x80);
		if(FreeSpace==0)
			return FT_VI800A_FIFO_Is_Empty;
		if(maxBytesToRead == 0)
			return FT_VI800A_Parameter_Value_Error;
		uint8_t AmountToRead = maxBytesToRead > FreeSpace ? FreeSpace : maxBytesToRead;		
		digitalWrite(sspin,LOW);
		SPI.transfer(address);  
		for(uint8_t i=0; i<AmountToRead; i++){
			array[i] = SPI.transfer(0x00);
		}
		digitalWrite(sspin,HIGH);
		numOfBytesRead = AmountToRead;
		return FT_VI800A_OK;
	}
	
	


	/*
	Description: Set the transmission data format.  This function can be called after library initialization to change the format for certain situations such as sending an address byte with a parity of 1 for the RS485 flow control.  It's not recommended to change the data format while data is being transmitted.  
	
	Parameters:
		parityType - The parity type of the data format, one in the FT_Data_Format_Parity_Type table.
		
		stopBitLength - The stop bit length of the data format, one in the FT_Data_Format_Stop_Bit_Length table.
		
		wordLength - The word length of the data format, one in the FT_Data_Format_Word_Length table.
		
	Return:
		None
	*/
	void FT_VI800A::SetDataFormat(FT_VI800A_Parity_Type parityType, FT_VI800A_Stop_Bit_Length stopBitLength, FT_VI800A_Word_Length wordLength){
		WriteToRegister(FT_VI800A_LCR, (ReadRegister(FT_VI800A_LCR)&FT_LCR_DATAFORMAT_MASK)|(uint8_t)parityType|(uint8_t)stopBitLength|(uint8_t)wordLength);
	}
	
	/*
	Description: Set the data transmission baud rate.  Setting the baud rate should not be done after enabling the sleep mode.  The baud rate is also being set by the Init() function.
	
	Parameter:
		baudrate - the desired baud rate.  Valid integers are between 15 and 921600, inclusive. Refer to the one of the Init() function for the limitation of the baud rate value.
		
	Return: None.
	*/
	void FT_VI800A::SetBaudRate(uint32_t baudrate){
		uint8_t DataFormat = ReadRegister(FT_VI800A_LCR);
		uint16_t divisor;
		WriteToRegister(FT_VI800A_LCR,FT_LCR_ENABLE_DIVISOR_LATCH); 
		divisor=FT_VI800A_MAX_FREQUENCY / (baudrate<<4); 
		WriteToRegister(FT_VI800A_DLH, divisor>>8); 
		WriteToRegister(FT_VI800A_DLL, divisor & 0xff); 
		WriteToRegister(FT_VI800A_LCR,DataFormat);
	}
	
	/*
	Description: Overloaded Init() function to initializes the device using default values of: Baud rate at 9600, no parity bit, one stop bit, and 8 data bits. This function calls the Init() with the mentioned default values.
	
	Parameters:
		SSpin - the add-on module slave select pin.
		
	Return: 
		FT_VI800A_Status - One of the value in the FT_VI800A_Status table.
	
	*/
	FT_VI800A_Status FT_VI800A::Init(uint8_t SSpin){
		return Init(SSpin, 9600,FT_VI800A_No_Parity, FT_VI800A_One_Stop_Bit, FT_VI800A_Eight_Data_Bits);
	}
	
	/*
	Description: This function initializes the Addon board with user specified slave select pin, desired baud rate, and the transmission data format.  The Init() function should be the first library function to be called after instantiated the library object.    
	
	Parameter:
		SSpin - The Addon module slave select pin.
		
		BaudRate - this is the desired communication baud rate, min: 15, max 921600.  NOTE: not all baud rates within the range are "supported" because of the rounding error on some "non-standard" baud rates.  The final baud rate uses the following equation to calculates a divisor for the internal baud rate generator logic unit: divisor = 14745600 / (baud rate*16) . The 16-bit divisor can only take integer values, so if the division did not yield an integer value then the actual baud rate will be a little different than the desired rate.
		
		parityType - The parity type of the data format, one in the FT_Data_Format_Parity_Type table.
		
		stopBitLength - The stop bit length of the data format, one in the FT_Data_Format_Stop_Bit_Length table.
		
		wordLength - The word length of the data format, one in the FT_Data_Format_Word_Length table.
	
	Return:
		One in the FT_VI800A_Status table.
	*/
	FT_VI800A_Status FT_VI800A::Init(uint8_t SSpin, uint32_t BaudRate, FT_VI800A_Parity_Type parityType, FT_VI800A_Stop_Bit_Length stopBitLength, FT_VI800A_Word_Length wordLength){
		uint16_t divisor=0;
		FCR=0;
		if(BaudRate<FT_VI800A_MIN_BAUD_RATE || BaudRate >FT_VI800A_MAX_BAUD_RATE){
			return FT_VI800A_Configuration_Error;
		}

		pinMode(SSpin,OUTPUT);
		pinMode(A2,OUTPUT);  
		digitalWrite(A2,HIGH); 
		digitalWrite(SSpin,HIGH);
		
		SPI.begin();
		/* For Due, Yun, Galileo and other hardware */
		#if defined(__SAM3X8E__)
			SPI.setClockDivider(4);//hard coding for due board
		#else	
			/* Other Arduino boards */
			SPI.setClockDivider(SPI_CLOCK_DIV2);//for Intel Galileo this value should be fine
		#endif	

		SPI.setBitOrder(MSBFIRST);
		SPI.setDataMode(SPI_MODE0);
		
		sspin=SSpin;
			

		//setting the baud rate and the registers are only accessible when LCR[7]=1
		WriteToRegister(FT_VI800A_LCR,FT_LCR_ENABLE_DIVISOR_LATCH); //The baud rate register is only accessible when LCR[7]=1
		divisor=FT_VI800A_MAX_FREQUENCY / (BaudRate<<4); //calculate the baud rate divisor for the internal baud rate generation logic unit
		WriteToRegister(FT_VI800A_DLH, divisor>>8); //DLH holds the upper 8 bits of the divisor
		WriteToRegister(FT_VI800A_DLL, divisor & 0xff); //DLL holds the lower 8 bits of the divisor
		//Enable enhanced features by default
		WriteToRegister(FT_VI800A_LCR, FT_LCR_ACCESS_ENHANCED_REGISTERS);
		WriteToRegister(FT_VI800A_EFR, FT_EFR_ENABLE_ENHANCED_FUNCTIONS);
		WriteToRegister(FT_VI800A_LCR, (uint8_t)parityType|(uint8_t)stopBitLength|(uint8_t)wordLength);
		return FT_VI800A_OK;
	}
	
	/*
	Description: This function reports the current free space in the transmit FIFO.  The function automatically adjust the number to reflect whether the FIFO is enabled or not.  If FIFO is not enabled then a maximum of 1 free byte will get reported, otherwise the reported value can be any number up to 64.

	Parameters:
		None
		
	Return:
		An 8-bit value between 0 to 64
	*/
	uint8_t FT_VI800A::GetTransmitFIFOFreeSpace(){
		uint8_t FIFOlevel = ReadRegister(FT_VI800A_TXLVL);
		if((FCR&FT_FCR_ENABLE_TX_RX_FIFO) == FT_FCR_ENABLE_TX_RX_FIFO)
			return FIFOlevel;
		else
			return FIFOlevel==FT_VI800A_MAX_FIFO? 1 : 0;

	}
	
	/*
	Description: This function reports the current fill level, amount of occupying data, in the receive FIFO.  NOTE: The number reported from this function is based on the whole 64 byte FIFO, regardless of whether FIFO is enabled or not.  If FIFO is not enabled then the FIFO is restricted to one character.  
	
	Parameters: 
		This function reports the current fill level, amount of occupying space, in the receive FIFO.
		
	Return:
		An 8-bit value between 0 to 64
	*/
	uint8_t FT_VI800A::GetReceiveFIFOFillLevel(){
		return ReadRegister(FT_VI800A_RXLVL);
	}
	
	
	/*
	Description: This function enables the specified interrupts to the IER register.  Any interrupts that are not part of the parameter are disabled.
	
	Parameters: 
		Interrupts - one or more in the FT_VI800A_Enableable_Interrupts table.
		
	Return:
		None
	*/
	void FT_VI800A::SetInterrupts(uint8_t interrupts){
		WriteToRegister(FT_VI800A_IER, interrupts);
	}
	
	/*
	Description: This function can enable or disable sleep mode.  If sleep mode is enabled the device will greatly reduces its the power consumption but there are multiple implications while the device is in sleep mode. First, baud rate can not be set while the device is in sleep mode. 
	
	Parameters:
		FT_VI800A_OPTIONS option - One in the FT_VI800A_OPTIONS table, valid values are either FT_VI800A_Enable to enable sleep mode or FT_VI800A_Disable to disable sleep mode.
		
	Return:
		None.
	*/
	void FT_VI800A::EnableSleepMode(FT_VI800A_OPTIONS option){
		if(option == FT_VI800A_Enable)
			WriteToRegister(FT_VI800A_IER,ReadRegister(FT_VI800A_IER)| FT_IER_SLEEP_MODE);//IER[4]=1 to enable sleep mode
		else if(option == FT_VI800A_Disable)
			WriteToRegister(FT_VI800A_IER,ReadRegister(FT_VI800A_IER) & ~FT_IER_SLEEP_MODE);//IER[4]=0 to disable sleep mode
	}
	
	/*
	Description: Perform a software reset.  All device registers will reset to their default value after calling this function.  After calling this function, the application should reset the transmission data format, SetDataFormat().  The baud rate, scratch pad register, and special characters hold their initialized values during the reset.
	
	Parameters: 
		None
	
	Return: 
		None
	*/
	void FT_VI800A::SWReset(){
		WriteToRegister(FT_VI800A_IOCTRL, FT_IOCONTROL_SOFTWARE_RESET);
		WriteToRegister(FT_VI800A_FCR, 0x00);
		FCR=0;
	}
	
	
	
	/*
	Description: The severity of the interrupt is same as order in the entries of the FT_VI800A_Interrupts_Status table.  Modem interrupt status must be read XXX function.  GPIO interrupt must check the GPIO state from the GetIOPinDirection() function against the previous known state. Application should compare the entries in the order listed in the FT_VI800A_Interrupts_Status table, only the highest priority interrupt will get reported.
	
	Parameters: 
		None
	
	Return: 
		One in the FT_VI800A_Interrupts_Status table.
	*/
	uint8_t FT_VI800A::GetInterrupts(void){
		//one of FT_IIR_Interrupts_Status 
		uint8_t status = ReadRegister(FT_VI800A_IIR) & FT_IIR_ALL_INTERRUPTS;
			if(status == FT_VI800A_Receiver_Line_Status_Error)
				return FT_VI800A_Receiver_Line_Status_Error;

			else if(status== FT_VI800A_Receiver_Time_Out_Interrupt)
				return FT_VI800A_Receiver_Time_Out_Interrupt;

			else if(status == FT_VI800A_RHR_Interrupt)
				return FT_VI800A_RHR_Interrupt;

			else if(status == FT_VI800A_THR_Interrupt)
				return FT_VI800A_THR_Interrupt;
				
			else if(status == FT_VI800A_Modem_Interrupt)
				return FT_VI800A_Modem_Interrupt;

			else if(status == FT_VI800A_Input_Pin_State_Change)
				return FT_VI800A_Input_Pin_State_Change;
				
			else if(status == FT_VI800A_Received_Xoff_Signal)
				return FT_VI800A_Received_Xoff_Signal;
				
			else if(status == FT_VI800A_CTS_RTS_Inactive)
				return FT_VI800A_CTS_RTS_Inactive;
			else
				return FT_VI800A_No_Interrupts;		
	}
	
	/*
	bool FT_VI800A::CheckInterrupts(FT_VI800A_Interrupts_Status interrupts){
		//check one or more interrupts because at times the other interrupts might not be usefull to the current enabled feature.
			if((Read(FT_VI800A_IIR) & interrupts) == interrupts){
				return true;
			}
			else{
				return false;
			}
	}
*/	
	
	
	

	
/*
Description: This function sets the desired flow control and its required parameters.  Optional features such as FIFO, enable interrupts, are not being set in this function but they should be called before this function.

Parameters: 

FlowControlType(one in the FT_VI800A_Flow_Type table):
FT_VI800A_No_Flow : not enhanced flow control, pure uart mode, one byte buffer.
FT_VI800A_SW_Flow for enhanced software flow control.
FT_VI800A_HW_Flow for enhanced hardware flow control.
FT_VI800A_RS485_Flow for enhanced auto rs485 with optional slave address detection.

If FT_VI800A_SW_FLOW is used:
param1: one group in the FT_VI800A_SW_Flow_Combination table.
param2: XON1 - Special character to signal start transfer, even not being used.
param3: XON2 - Special character to signal start transfer, even not being used.
param4: XOFF1 - Special character to signal stop transfer, even not being used.
param5: XOFF2 - Special character to signal stop transfer, even not being used.
param6: one in the FT_VI800A_Flow_Trigger table for the FIFO trigger level
param7: if FT_VI800A_Use_TCR_RX_Trigger is specified in param6 then this value represents the halt trigger level, if FT_VI800A_Use_FCR_RX_Trigger is specified then this parameter takes one in the FT_VI800A_Preset_RX_FIFO_Trigger_Level table.
param8: if FT_VI800A_Use_TCR_RX_Trigger is specified in param6 then this value represents the resume trigger level.


For hardware flow control:
param2: (Optional parameter) RX halt trigger level.  Alternatively, the trigger level can be set by SetReceiveFIFOTriggerLevel() function but it has to be set before calling this function.  If the trigger level was set previously then it doesn't need to be set again, except new values. NOTE: the halt trigger level has to be greater than the resume level, please see the SetReceiveFIFOTriggerLevel function description for more information on the requirements of the value.
param3: (OPtional parameter) RX resume trigger level. Alternatively, the trigger level can be set by SetReceiveFIFOTriggerLevel function but it has to be set before calling this function.  If the trigger level was set previously then it doesn't need to be set again, except new values.  NOTE: the resume trigger level has to be less than the resume level, please see the SetReceiveFIFOTriggerLevel function description for for more information on the requirements of the value.

For Auto RS485:
param1: one or more in the FT_VI800A_RS485_FLOW_OPTIONS table
param2: slave address (Only if FT_Auto_RS485_Address_Detection option was used, otherwise this parameter can be omitted)
*/

FT_VI800A_Status FT_VI800A::SetFlowControl(FT_VI800A_Flow_Type FlowControlType, uint8_t param1=0, uint8_t param2=0, uint8_t param3=0, uint8_t param4=0, uint8_t param5=0, uint8_t param6=0, uint8_t param7=0, uint8_t param8=0){
	uint8_t DataFormat = ReadRegister(FT_VI800A_LCR); //

	ResetFlowFunctions();
	if(FlowControlType == FT_VI800A_No_Flow){
#ifdef VI800A_N485U_MODULE
WriteToRegister(FT_VI800A_EFCR, ReadRegister(FT_VI800A_EFCR)|FT_EFCR_TRANSMITTER_CONTROSL_RTS);
#endif
		return FT_VI800A_OK;
	}
	else if(FlowControlType==FT_VI800A_SW_Flow){
		if(param6==FT_VI800A_Use_TCR_RX_Trigger){
			if(param7==0 || param7<param8 || param7&0x03 || param8&0x03)
				return FT_VI800A_Parameter_Value_Error;
			/*
			WriteToRegister(FT_VI800A_MCR, ReadRegister(FT_VI800A_MCR)|FT_MCR_ENABLE_TCR_TLR_REGISTERS);
			FCR =  (FCR&(~FT_FCR_RX_TRIGGER_LEVEL)) | FT_FCR_ENABLE_TX_RX_FIFO | FT_VI800A_RX_Sixty_Characters;
			WriteToRegister(FT_VI800A_FCR, FCR);
			WriteToRegister(FT_VI800A_TCR, (param8<<2)|(param7>>2));
			*/
			WriteToRegister(FT_VI800A_MCR, ReadRegister(FT_VI800A_MCR)|FT_MCR_ENABLE_TCR_TLR_REGISTERS);
			WriteToRegister(FT_VI800A_TCR, (param8<<2)|(param7>>2));
			WriteToRegister(FT_VI800A_MCR, ReadRegister(FT_VI800A_MCR)& (~FT_MCR_ENABLE_TCR_TLR_REGISTERS));
			FCR =  (FCR&(~FT_FCR_RX_TRIGGER_LEVEL)) | FT_FCR_ENABLE_TX_RX_FIFO;
			WriteToRegister(FT_VI800A_FCR, FCR);
		}
		else if(param6==FT_VI800A_Use_FCR_RX_Trigger){
			WriteToRegister(FT_VI800A_MCR, ReadRegister(FT_VI800A_MCR)|FT_MCR_ENABLE_TCR_TLR_REGISTERS);
			WriteToRegister(FT_VI800A_TCR, 0x00);
			WriteToRegister(FT_VI800A_MCR, ReadRegister(FT_VI800A_MCR)& (~FT_MCR_ENABLE_TCR_TLR_REGISTERS));
			FCR = ((FCR&(~FT_FCR_RX_TRIGGER_LEVEL))|FT_FCR_ENABLE_TX_RX_FIFO)|param7;
			WriteToRegister(FT_VI800A_FCR, FCR);  //enable FIFO implicitly 
		}
		else if(param6==FT_VI800A_Use_No_RX_Trigger){
		}
		WriteToRegister(FT_VI800A_LCR, FT_LCR_ACCESS_ENHANCED_REGISTERS); 
		WriteToRegister(FT_VI800A_XON1,param2); 
		WriteToRegister(FT_VI800A_XON2,param3);
		WriteToRegister(FT_VI800A_XOFF1,param4);
		WriteToRegister(FT_VI800A_XOFF2,param5);
		WriteToRegister(FT_VI800A_EFR, (ReadRegister(FT_VI800A_EFR)&FT_EFR_SW_FLOW_MASK)|param1);  //software flow control activation is done as the last step.
		WriteToRegister(FT_VI800A_LCR,DataFormat);
		return FT_VI800A_OK;
	}
	else if(FlowControlType == FT_VI800A_HW_Flow){
		if(param1==FT_VI800A_Use_TCR_RX_Trigger){ //use TCR register for the FIFO trigger level
			if((param2==0) || (param2<param3) || param2&0x03 || param3&0x03)
				return FT_VI800A_Parameter_Value_Error;
			WriteToRegister(FT_VI800A_MCR, ReadRegister(FT_VI800A_MCR)|FT_MCR_ENABLE_TCR_TLR_REGISTERS);
			FCR |= (FT_FCR_ENABLE_TX_RX_FIFO);
			WriteToRegister(FT_VI800A_FCR, FCR);
			WriteToRegister(FT_VI800A_TCR, (param3<<2)|(param2>>2));
		}
		else if(param1==FT_VI800A_Use_FCR_RX_Trigger){ 
			WriteToRegister(FT_VI800A_MCR, ReadRegister(FT_VI800A_MCR)|FT_MCR_ENABLE_TCR_TLR_REGISTERS);
			WriteToRegister(FT_VI800A_TCR, 0x00);
			WriteToRegister(FT_VI800A_MCR, ReadRegister(FT_VI800A_MCR)& (~FT_MCR_ENABLE_TCR_TLR_REGISTERS));
			FCR = (FCR & (~FT_FCR_RX_TRIGGER_LEVEL)) | FT_FCR_ENABLE_TX_RX_FIFO | param2;
			WriteToRegister(FT_VI800A_FCR, FCR);
		}
		else if(param1 == FT_VI800A_Use_No_RX_Trigger){
		
		}
		WriteToRegister(FT_VI800A_LCR,FT_LCR_ACCESS_ENHANCED_REGISTERS);
		WriteToRegister(FT_VI800A_EFR,ReadRegister(FT_VI800A_EFR)|FT_EFR_ENABLE_CTS_FLOW_CONTROL|FT_EFR_ENABLE_RTS_FLOW_CONTROL); 
		WriteToRegister(FT_VI800A_LCR,DataFormat);
		return FT_VI800A_OK;
	}
	else if(FlowControlType == FT_VI800A_RS485_Flow){
		WriteToRegister(FT_VI800A_EFCR, ReadRegister(FT_VI800A_EFCR)|FT_EFCR_TRANSMITTER_CONTROSL_RTS);
		
		if((param1&FT_VI800A_Auto_RS485_Address_Detection_Mode) == FT_VI800A_Auto_RS485_Address_Detection_Mode){
			WriteToRegister(FT_VI800A_LCR, FT_LCR_ACCESS_ENHANCED_REGISTERS);
			WriteToRegister(FT_VI800A_EFR,ReadRegister(FT_VI800A_EFR)|FT_EFR_ENABLE_SPECIAL_CHAR_DETECTION); 
			//this sets the xoff2 special character recognition.  If a match occurs, the received data is transferred to FIFO and the special character interrupt will be raised to indicate a special character has been detected, if the special character interrupt was enabled.  This feature is used by auto Rs485 with slave address detection.
			WriteToRegister(FT_VI800A_XOFF2,param2); 
			WriteToRegister(FT_VI800A_LCR,DataFormat);
		}
		else if((param1&FT_VI800A_RS485_Multidrop_Mode) == FT_VI800A_RS485_Multidrop_Mode){
		}
		
		EnableReceiver(FT_VI800A_Disable);
		
		//RS485 mode requires the parity type to be zero and setting the parity type to anything else will prevent the device to detect the address byte.  
		WriteToRegister(FT_VI800A_EFCR, ReadRegister(FT_VI800A_EFCR)|FT_EFCR_ENABLE_RS485_MODE);//Enable RS485 flow.		
		return FT_VI800A_OK;
	}
	else{
		return FT_VI800A_Configuration_Error;
	}
}

/*
Description: Internal function to disable all flow control before setting the desired flow control with specific parameters. Software flow control, hardware flow control, and Auto RS485 control are mutually exclusive so only one can be enabled at any given time so the other enhanced features are required to be disabled. This function also attempts to enable the receive FIFO because it's disabled for RS485 flow control.

Parameters: 
	None
	
Return:
	None
*/
	void FT_VI800A::ResetFlowFunctions(){
		uint8_t DataFormat = ReadRegister(FT_VI800A_LCR);
		WriteToRegister(FT_VI800A_LCR, FT_LCR_ACCESS_ENHANCED_REGISTERS);
		WriteToRegister(FT_VI800A_EFR, FT_EFR_ENABLE_ENHANCED_FUNCTIONS); 
		WriteToRegister(FT_VI800A_LCR, DataFormat);
		WriteToRegister(FT_VI800A_EFCR, ReadRegister(FT_VI800A_EFCR) & FT_EFCR_DISABLE_RS485_FEATURES);
		WriteToRegister(FT_VI800A_MCR, ReadRegister(FT_VI800A_MCR)&(~FT_MCR_ENABLE_TCR_TLR_REGISTERS));
	}
	

/*
Description: Enable or disable the the transmit and receive FIFO.  By default, the FIFO is not enabled after calling Init() or reset the device.

Parameters:
	FT_VI800A_OPTIONS option - one in the FT_VI800A_OPTIONS table, valid values are either FT_VI800A_Enable to enable the FIFO or FT_VI800A_Disable to disable the FIFO.
	
Return:
	None.
*/
	void FT_VI800A::EnableFIFO(FT_VI800A_OPTIONS option){
		if(option == FT_VI800A_Enable){
			FCR |= FT_FCR_ENABLE_TX_RX_FIFO;
			WriteToRegister(FT_VI800A_FCR,FCR);
		}
		else if(option == FT_VI800A_Disable){
			FCR &= (~FT_FCR_ENABLE_TX_RX_FIFO);
			WriteToRegister(FT_VI800A_FCR,FCR);
		}
	}
	
	/*
	Description: disable/enable the transmitter so UART will not send serial data out on the transmit line, the transmitter is enabled by default.  If the receiver is not disabled then it will continue to receive data.  If there is data in the transmit FIFO after calling this function, the data will get sent out before the transmitter goes into disable state.  
	
	Parameter: 
		option - one in the FT_VI800A_OPTIONS table, valid values are FT_VI800A_Disable to disable the transmitter and FT_VI800A_Enable to enable the transmitter.
		
	Return:
		None.
	*/
	void FT_VI800A::EnableTransmitter(FT_VI800A_OPTIONS option){
		if(option == FT_VI800A_Disable)
			WriteToRegister(FT_VI800A_EFCR,ReadRegister(FT_VI800A_EFCR)|FT_EFCR_DISABLE_TRANSMITTER);
		else if(option == FT_VI800A_Enable)
			WriteToRegister(FT_VI800A_EFCR,ReadRegister(FT_VI800A_EFCR)& ~(FT_EFCR_DISABLE_TRANSMITTER));
	}
	
	
	/*
	Description: Disable/enable only the receiver so UART will not receive any more data, the receiver is enabled by default.  It's not recommended to disable the receiver during data receiving. NOTE: Auto RS485 mode enables/disables the receiver automatically.
	
	Parameter: 
		option - one in the FT_VI800A_OPTIONS table, valid values are FT_VI800A_Disable to disable the receiver and FT_VI800A_Enable to enable the receiver.
		
	Return:
		None.
	*/
	void FT_VI800A::EnableReceiver(FT_VI800A_OPTIONS option){
		if(option == FT_VI800A_Disable)
			WriteToRegister(FT_VI800A_EFCR,ReadRegister(FT_VI800A_EFCR)|FT_EFCR_DISABLE_RECEIVER);
		else if(option == FT_VI800A_Enable)
			WriteToRegister(FT_VI800A_EFCR,ReadRegister(FT_VI800A_EFCR)& (~FT_EFCR_DISABLE_RECEIVER));
	}
	
/*
Description: Get all the error in the receive FIFO, all FIFO errors have a non-zero value.  The FT_VI800A_FIFO_Status table has all the errors that can occur in the receive FIFO.  Errors in a character are identified by calling this function then read the next character in the FIFO because break interrupt, framing error, and parity error are indicated on the top of the RX FIFO(next character to be read). Polling errors in the FIFO should not enable the following interrupts: FT_VI800A_Modem_Status_Interrupt, FT_VI800A_Receive_Line_Status_Interrupt, FT_VI800A_Transmit_Holding_Register_Interrupt, and  FT_VI800A_Receive_Holding_Register_Interrupt.   

Parameters:
	None.
	
Return:
	an 8-bit value of all the receive FIFO errors.
*/
	uint8_t FT_VI800A::GetReceiveFIFOStatus(void){
		return (ReadRegister(FT_VI800A_LSR) & FT_LSR_ALL_FIFO_ERRORS);
	}
	
	/* 
	//Set the special character, this function assumes the LCR register has been set to access this register.
	void FT_VI800A::SetXON1SpecialCharacter(uint8_t xon1){
		Write(FT_VI800A_XON1,xon1);
	}
	
	// Set the special character, this function assumes the LCR register has been set to access this register.
	void FT_VI800A::SetXON2SpecialCharacter(uint8_t xon2){
		Write(FT_VI800A_XON2,xon2);
	}
	
	// Set the special character, this function assumes the LCR register has been set to access this register.
	void FT_VI800A::SetXOFF1SpecialCharacter(uint8_t xoff1){
		Write(FT_VI800A_XOFF1,xoff1);
	}

	// Set the special character, this function assumes the LCR register has been set to access this register.
	void FT_VI800A::SetXOFF2SpecialCharacter(uint8_t xoff2){
		Write(FT_VI800A_XOFF2,xoff2);
	}
	*/

	/*
	Description: Helper API to send the XON1 special character to the receiver. The enhanced auto RS485 requires the address byte to have a parity bit of 1 to the receiver.
	
	Parameters: 
		None.
		
	Return:
		None.
	*/
	void FT_VI800A::SendXON1SpecialCharacter(){
		uint8_t DataFormat = ReadRegister(FT_VI800A_LCR);
		WriteToRegister(FT_VI800A_LCR, FT_LCR_ACCESS_ENHANCED_REGISTERS);  
		Write(ReadRegister(FT_VI800A_XON1));
		WriteToRegister(FT_VI800A_LCR, DataFormat);
	}
	
	/*
	Description: Helper API to send the XON2 special character to the receiver.  The enhanced auto RS485 requires the address byte to have a parity bit of 1 to the receiver.
	
	Parameters: 
		None.
		
	Return:
		None.
	*/
	void FT_VI800A::SendXON2SpecialCharacter(){
		uint8_t DataFormat = ReadRegister(FT_VI800A_LCR);
		WriteToRegister(FT_VI800A_LCR, FT_LCR_ACCESS_ENHANCED_REGISTERS);  
		Write(ReadRegister(FT_VI800A_XON2));
		WriteToRegister(FT_VI800A_LCR, DataFormat);
	}

	/*
	Description: Helper API to send the XOFF1 special character to the receiver.  The enhanced auto RS485 requires the address byte to have a parity bit of 1 to the receiver.
	
	Parameters: 
		None.
		
	Return:
		None.
	*/
	void FT_VI800A::SendXOFF1SpecialCharacter(){
		uint8_t DataFormat = ReadRegister(FT_VI800A_LCR);
		WriteToRegister(FT_VI800A_LCR, FT_LCR_ACCESS_ENHANCED_REGISTERS);  
		Write(ReadRegister(FT_VI800A_XOFF1));
		WriteToRegister(FT_VI800A_LCR, DataFormat);
	}
	
	/*
	Description: Helper API to send the XOFF2 special character to the receiver.  The enhanced auto RS485 requires the address byte to have a parity bit of 1 to the receiver.
	
	Parameters: 
		None.
		
	Return:
		None.
	*/
	void FT_VI800A::SendXOFF2SpecialCharacter(){
		uint8_t DataFormat = ReadRegister(FT_VI800A_LCR);
		WriteToRegister(FT_VI800A_LCR, FT_LCR_ACCESS_ENHANCED_REGISTERS);  
		Write(ReadRegister(FT_VI800A_XOFF2));
		WriteToRegister(FT_VI800A_LCR, DataFormat);
	}
	
	/*
	Description: Manually enable/disable the RTS output.  This function should only be used for manual CTS/RTS flow control.  Internal loop back and enhanced hardware flow control automatically handle the CTS and RTS lines.
	
	Parameters:
		FT_VI800A_Enable in the FT_VI800A_OPTIONS table to force RTS output to active.
		FT_VI800A_Disable in the FT_VI800A_OPTIONS table to force RTS output to inactive.
		
	Return:
		None.
	*/
	void FT_VI800A::EnableRTSOutput(FT_VI800A_OPTIONS option){
		if(option == FT_VI800A_Enable){
			WriteToRegister(FT_VI800A_MCR, ReadRegister(FT_VI800A_MCR)|FT_MCR_FORCE_RTS_TO_ACTIVE);
		}
		else if(option == FT_VI800A_Disable){
			WriteToRegister(FT_VI800A_MCR, (ReadRegister(FT_VI800A_MCR)&(~FT_MCR_FORCE_RTS_TO_ACTIVE)));
		}
	}
	
	/*
	Description: Check the status of the CTS line.  TCR register FIFO trigger levels should not be used for the whole duration while this function is being utilized for manual CTS/RTS flow control. This function should only be used with EnableRTSOutput to manually data flow control using the CTS and RTS lines.  Internal loop back and enhanced hardware flow control automatically handle the CTS and RTS lines.
	
	Parameters:
		None
		
	Return:
		boolean true - CTS is active as an indication that data transmission can occur.
		boolean false - CTS is inactive as an indication that data transmission should not be made.
	*/
	bool FT_VI800A::isCTSActive(void){
		if(ReadRegister(FT_VI800A_MSR)&FT_MSR_COMPLEMENT_OF_THE_CTS_INPUT)
			return true;
		else
			return false;
	}
	

	/*
	This function sets the transmit and receive FIFO trigger levels used for interrupt generation.  If one of the trigger level is 0 then the trigger level specified by the SetPresetFIFOInterruptTriggerLevel function will be used instead and there are only 4 fixed trigger.  
	
	Parameter:
		ReceiveFIFOInterruptLevel: uint8_t RXFIFOTriggerLevel - an 8-bit value representing the trigger level for the receive FIFO.  If the FIFO is not enabled then the RHR interrupt is reporting the data in the receive FIFO is ready to read.  If the FIFO is enabled then the RHR interrupt is reporting when the receive FIFO is above the specified trigger level. Valid values are from 4 to 60, with a granularity of 4.
		TransmitFIFOInterruptLevel: an 8-bit value representing the trigger level for the transmit FIFO.  If the FIFO is not enabled then the THR interrupt is reporting the transmit FIFO is empty.  If the FIFO is enabled then the THR interrupt is reporting when the transmit FIFO is above the specified trigger level. Valid values are from 4 to 60, with a granularity of 4.
	*/
	FT_VI800A_Status FT_VI800A::SetFIFOInterruptTriggerLevel(uint8_t ReceiveFIFOInterruptLevel, uint8_t TransmitFIFOInterruptLevel){
		if(TransmitFIFOInterruptLevel < FT_VI800A_MIN_FIFO_TRIGGER_LEVEL || 
			TransmitFIFOInterruptLevel > FT_VI800A_MAX_FIFO_TRIGGER_LEVEL ||
			ReceiveFIFOInterruptLevel < FT_VI800A_MIN_FIFO_TRIGGER_LEVEL || 
			ReceiveFIFOInterruptLevel > FT_VI800A_MAX_FIFO_TRIGGER_LEVEL
			)
			return FT_VI800A_Configuration_Error;
		FCR &= FT_FCR_Trigger_Mask;
		WriteToRegister(FT_VI800A_FCR, FCR);
		WriteToRegister(FT_VI800A_MCR, ReadRegister(FT_VI800A_MCR)|FT_MCR_ENABLE_TCR_TLR_REGISTERS);
		WriteToRegister(FT_VI800A_TLR,(ReceiveFIFOInterruptLevel<<2) | (TransmitFIFOInterruptLevel>>2));
		WriteToRegister(FT_VI800A_MCR, ReadRegister(FT_VI800A_MCR)& (~FT_MCR_ENABLE_TCR_TLR_REGISTERS));
		return FT_VI800A_OK;
	}
	
	/*
	this function sets the preset interrupt trigger levels.  This function should not be called if the enhanced software and hardware flow control use the preset receive trigger level because this function modifies the same register.
	
	Parameters:
	ReceiveFIFOInterruptLevel - the RX FIFO trigger level, one in the FT_VI800A_Preset_RX_FIFO_Trigger_Level table.
	TransmitFIFOInterruptLevel - the TX FIFO trigger level, one in the FT_VI800A_Preset_TX_FIFO_Trigger_Level table.
	*/
	void FT_VI800A::SetPresetFIFOInterruptTriggerLevel(FT_VI800A_Preset_RX_FIFO_Trigger_Level ReceiveFIFOInterruptLevel, FT_VI800A_Preset_TX_FIFO_Trigger_Level TransmitFIFOInterruptLevel){
		WriteToRegister(FT_VI800A_MCR, ReadRegister(FT_VI800A_MCR)|FT_MCR_ENABLE_TCR_TLR_REGISTERS);
		WriteToRegister(FT_VI800A_TLR,0x00);
		WriteToRegister(FT_VI800A_MCR, ReadRegister(FT_VI800A_MCR)& (~FT_MCR_ENABLE_TCR_TLR_REGISTERS));
		FCR = (FCR&(0x0F))|(uint8_t)ReceiveFIFOInterruptLevel|(uint8_t)TransmitFIFOInterruptLevel;
		WriteToRegister(FT_VI800A_FCR, FCR);
	}
	

	


	

	/************************************
	Helper APIs
	************************************/
	/*
	Description: Read the scratch pad register.  The scratch pad register can be used by the user for temporary data storage, the data is retained even after a software reset.  
	
	Parameters:
		None.
		
	Return:
		An 8-bit value of the scratch pad.
	*/
	uint8_t FT_VI800A::ReadScratchPad(void){
		return ReadRegister(FT_VI800A_SPR);
	}
	
	
	/*
	Description: Write to the scratch pad register.  The scratch pad register can be used by the user for temporary data storage, the data is retained even after a software reset.  
	
	Parameters:
		An 8-bit value to the scratch pad.
		
	Return:
		None
	*/
	void FT_VI800A::WriteToScratchPad(uint8_t value){
		WriteToRegister(FT_VI800A_SPR,value);
	}
	
	/*
	Description: Set the GPIO pin direction as either input or output.   The device has a fixed 4 output and 4 inputs so the value should always be 0xF0.

	Parameters:
		An 8-bit value representing the 8 GPIO ports, bit value of 0 is input and 1 is output.
		
	Return:
		None
	*/
	void FT_VI800A::SetIOPinDirection(uint8_t Direction){
		WriteToRegister(FT_VI800A_IODIR,Direction);
		
	}
	
	/*
	Description: Get the current the GPIO pin state.   
	
	Parameters:
		None.
		
	Return:
		An 8-bit value representing the current state of the GPIO pins, bit value of 0 indicates the pin number is set to as an input pin and bit value of 1 indicates the pin number is set to as an output pin.
	*/
	uint8_t FT_VI800A::GetIOPinDirection(void){
		return ReadRegister(FT_VI800A_IOSTATE);
	}
	

	/*
	Description: Set the GPIO pin state.  This function will only affect pins that are set to be output pins, setting input pins will not have any effects.
	
	Parameters:
		An 8-bit value representing the state of the output pins.  Bit value of 0 is setting the output pin number to zero and bit value of 1 is setting the output pin number to one.
		
	Return:
		None
	*/
	void FT_VI800A::SetIOPinState(uint8_t State){
		WriteToRegister(FT_VI800A_IOSTATE,State);
	}
	
	
	/*
	Description: Enable/disable the internal loop back mode.  If enabled then the transmitter will be connected to the receiver internally.  The internal loop back should not be enabled if the application intents to connect to other devices other than itself.  Internal loop back is disabled by default.
	
	Parameters:
		FT_VI800A_OPTIONS option - The valid values are in the FT_VI800A_OPTIONS table they're either FT_VI800A_Enable to enable internal loop back or FT_VI800A_Disable to disable internal loop back.
		
	Return:
		None
	*/
	void FT_VI800A::SetInternalLoopBack(FT_VI800A_OPTIONS option){
		if(option == FT_VI800A_Enable)
			WriteToRegister(FT_VI800A_MCR,ReadRegister(FT_VI800A_MCR)|FT_MCR_ENABLE_INTERNAL_LOOPBACK);
		else if(option == FT_VI800A_Disable)
			WriteToRegister(FT_VI800A_MCR,ReadRegister(FT_VI800A_MCR) & (~FT_MCR_ENABLE_INTERNAL_LOOPBACK));
	}
	
	
	/*
	void FT_VI800A::EnableModemPins(void){
		Write(FT_VI800A_IOCTRL, Read(FT_VI800A_IOCTRL) | FT_IOCONTROL_MODEM_GPIO);
	}
	*/
	
	/*
	Description: This function causes a break condition to be transmitted.  As long as the  StopTransmitLineBreak() function was not called the device will continuously send the line break condition to the receiver.  Line break condition is represented by a value of 0x00 that is on for a character time frame.
	
	Parameters:
		None.
		
	Return:
		None.
	*/
	void FT_VI800A::StartTransmitLineBreak(void){
		WriteToRegister(FT_VI800A_LCR, ReadRegister(FT_VI800A_LCR) | FT_LCR_LINE_BREAK_CONDITION);
	}
	
	/*
	Description: This function stops the transmission of the line break condition to the receiver by the StartTransmitLineBreak() function. It has no effect if line break is not being sent.
	
	Parameters:
		None.
		
	Return:
		None.
	*/
	void FT_VI800A::StopTransmitLineBreak(void){
		WriteToRegister(FT_VI800A_LCR, ReadRegister(FT_VI800A_LCR) & (~FT_LCR_LINE_BREAK_CONDITION));
	}
	

	
	/*
	Description: Check the receiver FIFO to see if there is available data to be read.  
	
	Parameters:
		None.
		
	Return:
		boolean value: true - if there is at least 1 byte in the fifo, false - if no data is in the FIFO
	*/
	bool FT_VI800A::isDataInReceiver(void){
		if(GetReceiveFIFOFillLevel()>0)
			return true;
		else 
			return false;
	}
	
	/*
	Description: Check to see if both of the transmit FIFO and its shift register are empty.  The FIFO holds the data to be sent, regardless of whether FIFO is enabled or not, and the shift register package the next byte of the transmit FIFO to the connection format and sends it out, if the transmitter is enabled. The write function to the FIFO returns the number of bytes written to the FIFO but it doesn't mean the data is sent.  User can continuously poll this function after writing some data to the FIFO to make sure the data has been sent.
	
	Parameters:
		None.
		
	Return:
		Status of both of the transmit FIFO as well as the transmit shift register.
	*/
	bool FT_VI800A::isTransmitFIFOandShiftRegisterEmpty(void){
	//this function checks both transmit FIFO and the shift register to see if they're empty.
		return (ReadRegister(FT_VI800A_LSR)&FT_LSR_THR_TSR_ARE_EMPTY);//Check LSR[6]
	}
	
	
	/*
	Description: Check to see if the transmit FIFO is empty or not.  Alternatively, the GetTransmitFIFOFreeSpace() function can also be used to check the emptiness of the FIFO.  
	
	Parameters: 
		None
		
	Return:
		a boolean value representing whether the transmit FIFO is empty or not.
	*/
	bool FT_VI800A::isTransmitFIFOEmpty(void){
		return ReadRegister(FT_VI800A_LSR)&FT_LSR_THR_IS_EMPTY;//check LSR[5]
	}
	
	/*
	Description: Clear the transmit and receive FIFO.  The FIFO is not clear on soft reset and user can call this function to clear any data in the FIFO before using a new flow control.  
	
	Parameters: 
		None
		
	Return:
		None
	*/
	void FT_VI800A::clearFIFO(void){
		WriteToRegister(FT_VI800A_FCR, FCR|FT_FCR_RX_FIFO_RESET|FT_FCR_TX_FIFO_RESET); //the respective FIFO reset bit automatically resets itself after the process has been completed
	}
	
	
	
	
	/*
	Description: Ends the SPI.
	
	Parameters: 
		None
		
	Return:
		None
	*/
	void FT_VI800A::Exit(void){
		SPI.end();
	}
	
#endif
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	